/*
    Copyright (C) 2001-2002 by theKompany.com <www.thekompany.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    By PLUGINS we mean the commercial plug-ins developed by theKompany.com
    for Aethera.

    In addition, as a special exception, theKompany.com gives permission
    to link the code of this program with PLUGINS (or with
    modified versions of PLUGINS that use the same license as PLUGINS),
    and distribute linked combinations including the two.  You must obey
    the GNU General Public License in all respects for all of the code used
    other than PLUGINS.  If you modify this file, you may extend this
    exception to your version of the file, but you are not obligated to do so.
    If you do not wish to do so, delete this exception statement from your
    version.

    This license grants you the ability to use PLUGINS with Aethera only
    and may not be used outside of Aethera.
    See also http://www.thekompany.com/products/license.txt for details.
*/
/***************************************************************************
                remotemailfolder.cpp  -  the remote mail folder class
                             -------------------
    begin                : Mon Mar  5 16:16:00 EET 2001
    copyright            : (C) 2001 by theKompany (www.thekompany.com>
    author               : Eugen Constantinescu
    email                : eug@thekompany.com
 ***************************************************************************/

#include "config.h"

#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qdatetime.h>

#include <remotemailfolder.h>
#include <indexclass.h>
#include <servernotifier.h>
#include <accounts.h>
#include <accountmanager.h>
#include <messagedevice.h>
#include <messagedescriptor.h>
#include <messagefactory.h>
#include <headerclass.h>
#include <orb.h>
#include <imap4handler.h>
#include <iosynchandler.h>

#include <settings.h>
#include <qsettingsextra.h>

RemoteMailFolder::RemoteMailFolder(const QString &_storageDevice)
  : MailFolder(_storageDevice)
{
  setupFiles();
  config();
  loadIndex();
  loadStatus();
  uidJar=new IMAP4_UIDJar(getStorageDevice());
}

RemoteMailFolder::RemoteMailFolder(const QString &_storageDevice, const QString account, const QString mailbox)
 : MailFolder(_storageDevice)
{
  setupFiles();
  makeConfig(account, mailbox);
  loadIndex();
  loadStatus();
  uidJar=new IMAP4_UIDJar(getStorageDevice());
}

RemoteMailFolder::~RemoteMailFolder()
{
  QString idMail;
  
  // synch the uidJar
  for(UIDMap::iterator it=uidJar->rcvList.begin(); it!=uidJar->rcvList.end(); ++it)
  {
    idMail = (QString((*it).second.info.c_str()));
    idMail = idMail.mid( idMail.findRev('/') + 1 );
    IndexClass *idx = indexCollection[idMail];
    if( !idx )
    {
      uidJar->rcvList.erase(it);
    }
  }
  
  uidJar->saveLists();
  if( !saveStatus() )
    qDebug("RemoteMailFolder: Saving status failed!");
  
  // make a force synching before removing the folder
  sync();
  
  saveStatus();
  delete uidJar;
}

void RemoteMailFolder::setupFiles()
{
  headersFileName=getStorageDevice()+"/headers";
  configFileName=getStorageDevice()+"/config";
}

void RemoteMailFolder::config()
{
  QFile f( configFileName );
  f.open( IO_ReadOnly );
  QDataStream s( &f );
  s >> accountName;
  s >> mailboxName;
  f.close();

//  qDebug("Account=%s||Mailbox=%s", accountName.latin1(), mailboxName.latin1());
}

void RemoteMailFolder::makeConfig(QString account, QString mailbox)
{
  accountName=account;
  mailboxName=mailbox;

  QFile f( configFileName );
  f.open( IO_WriteOnly );
  QDataStream s( &f );
  s << accountName;
  s << mailboxName;
  f.close();

//  qDebug("Account=%s||Mailbox=%s", accountName.latin1(), mailboxName.latin1());
}

QString RemoteMailFolder::account() const
{
  return accountName;
}

QString RemoteMailFolder::mailbox() const
{
  return mailboxName;
}

QString RemoteMailFolder::getMessagesFileName() const
{
  return headersFileName;
}

QString RemoteMailFolder::createFullMessage(const QCString &text, const QCString &uid,
                                            const QDateTime &rcvtime, const QString &account,
                                            const MessageClass * /* parsedMessage */,
                                            bool bSync, const unsigned flags)
{
  if( uid.isEmpty() )
    return QString::null;

  // create index, set id and add to collection
  IndexClass * index = new IndexClass(this);
  if(!index)
  {
    return QString::null;
  }
  
  syncIndexID(index);
  QString newUUID = index->getID();
  indexCollection.insert(newUUID, index);

  MessageClass rfcMessage(text);

  // create message device
  MessageDevice *dev=new MessageDevice(index);

  // populate descriptor
  MessageDescriptor &descriptor=dev->getDescriptor();
  descriptor.load(rfcMessage);

  // add other data (account, uid, received time, index id etc)
  descriptor.account = account;
  descriptor.indexID = newUUID;
  descriptor.receivedDate=(QCString)DateClass(rcvtime);
  
  // UID
  QCString sUidValidity;
  sUidValidity.sprintf("%u", uidValidity());
  descriptor.uid = uid + "@" + sUidValidity;

  // debug
//  qDebug("remotemailfolder: created message with:");
//  qDebug("\taccount %s", (const char *)descriptor.account);
//  qDebug("\tuid %s", (const char *)descriptor.uid);
//  qDebug("\treceived time %s", (const char *)descriptor.receivedDate);
//  qDebug("\tindex id %s", (const char *)descriptor.indexID);

  // Update the uid file
  QString urlMessage=name()+"/"+index->getID();
  uidJar->rcvInsert(descriptor.uid, urlMessage.latin1());

  // write message id and unread mark into index
  index->setMessageID(descriptor.messageID);
  if( flags&MailFolder::Seen )
  {
    descriptor.status="Read";
    index->setUnreadMark(false);
  }
  else
  {
    descriptor.status="New";
    index->setUnreadMark(true);
  }

  // update folder stats
  if(index->getUnreadMark()) incrementUnread();

  // save message
  QFile dataFile(getMessagesFileName());
  dataFile.open(IO_WriteOnly|IO_Append);

  // write MBOX information
  QCString mboxInfo;
  mboxInfo="From - " + descriptor.receivedDate + "\r\n";
  dataFile.writeBlock((const char *)mboxInfo, mboxInfo.length());

  // write uniblock data (index is uniblock by default)
  index->setUniblockOffset(dataFile.at());
  index->setUniblockLength(text.length());
  index->setMultipartOnly(false);

  dataFile.writeBlock((const char *)text, index->getUniblockLength());
  dataFile.writeBlock("\r\n", 2);

  dataFile.close();

  // create index part list and update offsets
  for(unsigned int i=0;i<rfcMessage.partList.count();i++)
    index->addPart(new MimePart(*rfcMessage.partList.at(i)), true);

  // notify client or wait for a sync signal
  if( bSync )
  {
    syncMessages.append( dev );
    // force sync for a better memory management when
    // downloading big email boxes
    if( 300 <= syncMessages.count() )
      sync();
  }
  else
  {
//     // add index and descriptor to hierarchy
//     if(Settings::instance()->readBoolEntry("/Aethera/Threading/Enable", false))
//     {
//       reparentIndex(index);
//       crossReferenceIndexSet(index);
//     }

    // save descriptor (that will save the index as well)
    dev->saveDescriptor();
    // cleanup
    delete dev;

    ServerNotifier::thisInstance()->objectCreated(name() + "/" + newUUID);
  }

  return newUUID;
}

QString RemoteMailFolder::createMessage(const QCString &text, const QCString &uid,
                                        const QDateTime &rcvtime, const QString &account,
                                        const MessageClass * /* parsedMessage */,
                                        bool bSync, const unsigned flags)
{
  if( uid.isEmpty() )
    return QString::null;

  // create index, set id and add to collection
  IndexClass * index = new IndexClass(this);
  if(!index)
  {
    return QString::null;
  }
  
  syncIndexID(index);
  QString newUUID = index->getID();
  indexCollection.insert(newUUID, index);

  // rfc-parse message header
  HeaderClass rfcHeader((const char*)text);
  // build the message from the header
  MessageClass rfcMessage(rfcHeader);

  // create message device
  MessageDevice *dev = new MessageDevice(index);
  // populate descriptor
  MessageDescriptor &descriptor = dev->getDescriptor();
  descriptor.load(rfcMessage);

  // add other data (account, uid, received time, index id etc)
  descriptor.account = account;
  descriptor.indexID = newUUID;
  descriptor.receivedDate = (QCString)DateClass(rcvtime);
  
  // UID
  QCString sUidValidity;
  sUidValidity.sprintf("%u", uidValidity());
  descriptor.uid = uid + "@" + sUidValidity;

  // debug
//  qDebug("remotemailfolder: created message with:");
//  qDebug("\taccount %s", (const char *)descriptor.account);
//  qDebug("\tuid %s", (const char *)descriptor.uid);
//  qDebug("\treceived time %s", (const char *)descriptor.receivedDate);
//  qDebug("\tindex id %s", (const char *)descriptor.indexID);

  // Update the uid file
  QString urlMessage = name() + "/" + newUUID;
  uidJar->rcvInsert(descriptor.uid, urlMessage.latin1());

  // write message id and unread mark into index
  index->setMessageID(descriptor.messageID);
  if( flags&MailFolder::Seen )
  {
    descriptor.status="Read";
    index->setUnreadMark(false);
  }
  else
  {
    descriptor.status="New";
    index->setUnreadMark(true);
  }

  // update folder stats
  if(index->getUnreadMark()) incrementUnread();

  // save message
  QFile dataFile(getMessagesFileName());
  dataFile.open(IO_WriteOnly|IO_Append);

  // write MBOX information
  QCString mboxInfo;
  mboxInfo="From - " + descriptor.receivedDate + "\r\n";
  dataFile.writeBlock((const char *)mboxInfo, mboxInfo.length());

  // write uniblock data (index is uniblock by default)
  index->setUniblockOffset(dataFile.at());
  index->setUniblockLength(text.length());
  index->setMultipartOnly(false);

  dataFile.writeBlock((const char *)text, index->getUniblockLength());
  dataFile.writeBlock("\r\n", 2);

  dataFile.close();

  // create index part list and update offsets
  for(unsigned int i=0;i<rfcMessage.partList.count();i++)
    index->addPart(new MimePart(*rfcMessage.partList.at(i)), true);

  // notify client or wait for a sync signal
  if( bSync )
  {
    syncMessages.append( dev );
    // force sync for a better memory management when
    // downloading big email boxes
    if( 300 <= syncMessages.count() )
      sync();
  }
  else
  {
//     // add index and descriptor to hierarchy
//     if(Settings::instance()->readBoolEntry("/Aethera/Threading/Enable", false))
//     {
//       reparentIndex(index);
//       crossReferenceIndexSet(index);
//     }

    // save descriptor (that will save the index as well)
    dev->saveDescriptor();
    // cleanup
    delete dev;

    ServerNotifier::thisInstance()->objectCreated(name() + "/" + newUUID);
  }

  return newUUID;
}

bool RemoteMailFolder::updateMessage(const QString & text, const QString & indexURL)
{
  if( indexURL.isEmpty() || text.isEmpty() )
    return false;

  MessageClass rfcMessage( QCString(text.latin1()) );
  IndexClass *index = ObjectRequestBroker::thisInstance()->indexReference( indexURL );
  if( !index )
  {
    qDebug("RemoteMailFolder::updateMessage : no index reference!");
    return false;
  }

  // Eugen C. - unlock the mail, the update is done
  index->unlock();
  
  // create message device
  MessageDevice *dev = new MessageDevice(index);
  dev->loadDescriptor();
  MessageDescriptor &descriptor=dev->getDescriptor();
  descriptor.load(rfcMessage);
  
  // Eugen C. - set it as read
  if( index->getUnreadMark() )
    index->getParentFolder()->decrementUnread();
  
  index->setUnreadMark(false);
  descriptor.status = "Read";
  
//  debug
//  qDebug("remotemailfolder: changed message with:");
//  qDebug("\taccount %s", (const char *)descriptor.account);
//  qDebug("\tuid %s", (const char *)descriptor.uid);
//  qDebug("\treceived time %s", (const char *)descriptor.receivedDate);
//  qDebug("\tindex id %s", (const char *)descriptor.indexID);
  
  // save message
  QFile dataFile(getMessagesFileName());
  dataFile.open(IO_WriteOnly|IO_Append);
  
  // write MBOX information
  QCString mboxInfo;
  mboxInfo="From - " + descriptor.receivedDate + "\r\n";
  dataFile.writeBlock((const char *)mboxInfo, mboxInfo.length());
  
  // write uniblock data (index is uniblock by default)
  index->setUniblockOffset(dataFile.at());
  index->setUniblockLength(text.length());
  index->setMultipartOnly(false);
  
  dataFile.writeBlock((const char *)text, index->getUniblockLength());
  dataFile.writeBlock("\r\n", 2);
  
  dataFile.close();
  
  // create index part list and update offsets
  for(unsigned int i=0;i<rfcMessage.partList.count();i++)
    index->addPart(new MimePart(*rfcMessage.partList.at(i)), true);
  
  // save descriptor (that will save the index as well)
  dev->saveDescriptor();
  
  // cleanup
  delete dev;

  ServerNotifier::thisInstance()->objectChanged(name()+"/"+index->getID());

  return true;
}

void RemoteMailFolder::sync()
{
  IndexClass * index = 0L;
  
  if( !syncMessages.count() )
  {
//    qDebug("MailFolder::sync() : Nothing to sync!");
  }
  else
  {
    qDebug("Sync %u mails.", syncMessages.count());
    
    // save the descriptors and build the messages url list
    QStringList serverURLs;
    QFile dataFile( getDescriptorFileName() );
    if(dataFile.open(IO_WriteOnly|IO_Append))
    {
      QDataStream stream(&dataFile);
      for(QPtrListIterator<MessageDevice> it(syncMessages); it.current(); ++it)
      {
        MessageDescriptor &descriptor = it.current()->getDescriptor();
        index = it.current()->index();
        
        serverURLs.append(name()+"/"+ descriptor.indexID);
        
        index->setDescriptorOffset(dataFile.at());
        stream << descriptor;
        index->setDescriptorLength(dataFile.at() - index->getDescriptorOffset());
        
      }
      dataFile.close();

      // save the whole index once
      saveIndex();
    
      // this should delete the device (descriptors) too
      syncMessages.clear();
    
      // notify the UI for the changes
      ServerNotifier::thisInstance()->objectCreated(serverURLs);
    }
    else
      qDebug("MailFolder::sync() : the descriptors cannot be saved!");
    
//    // Delete the messages from trash list.
//    if( !uidJar->trashList.size() )
//    {
//      for(UIDMap::iterator it=uidJar->trashList.begin(); it!=uidJar->trashList.end(); ++it)
//      {
//        IndexClass *idx=ObjectRequestBroker::thisInstance()->indexReference(QString((*it).second.info.c_str()));
//        if( idx )
//          deleteMessage(idx);
//      }
//      uidJar->trashList.clear();
//    }

    //Save the status
    uidJar->saveLists();
    if( !saveStatus() )
      qDebug("Saving status failed!");
  }
}

void RemoteMailFolder::loadIndex()
{
  // debug
  qDebug("mailfolder: [%s] is reading index...", name().latin1());
  
  IndexClass *tindex;
  setUnread(0);
  setPruneIndexCount(0);

  QFile f(getIndexFileName());
  QDataStream stream(&f);

  // check index version number
  if(f.open(IO_ReadOnly))
  {
    QString idxExpectedVersion=QString::number(INDEX_VERSION);
    QString idxVersion;
    stream>>idxVersion;

    if(idxExpectedVersion!=idxVersion)
    {
      qDebug("Folder %s attempted to read an older version or invalid index\
          cache, discarding index...", (const char *)name());
    }
    else
    {
      while(!stream.atEnd())
      {
        tindex=new IndexClass(this);

        // read index
        stream>>(*tindex);

        // check to see if the index was already included
        IndexClass *pruneIndex;
        if((pruneIndex=indexCollection[tindex->getID()]))
        {
          // ...it was, increment the prune list
          incrementPruneIndexCount();

          // update folder stats
          if(pruneIndex->getUnreadMark()) decrementUnread();
        }
        else
        {
          // update the last index id
          unsigned long tlast=tindex->getID().mid(3).toULong();
          if(tlast>getLastIndexID())
            setLastIndexID(tlast+1);
        }
        // add index to collection
        indexCollection.replace(tindex->getID(), tindex);
        
        // debug
        // qDebug("localmailfolder: [loadIndex] loaded %s...", (const char *)tindex->getID());

        // update folder stats
        if(tindex->getUnreadMark()) incrementUnread();
      }
    }
    
    f.close();
  }
}

bool RemoteMailFolder::shouldExpunge()
{
  return false;
}

bool RemoteMailFolder::deleteMessage(IndexClass *idx)
{
  if(!idx || idx->getParentFolder()!=this)
    return false;

  // remove the mail from the IMAP folder too
  if( !IMAP4Handler::ref()->deleteMessage(*idx) )
    return false;

  return true;
}

bool RemoteMailFolder::deleteTrashMessages()
{
//  qDebug("RemoteMailFolder::deleteTrashMessages");
  
  IndexClass * idx = 0L;
  QStringList paths;
  
  for(UIDMap::iterator it=uidJar->trashList.begin();
      it!=uidJar->trashList.end(); ++it)
  {
    idx = ObjectRequestBroker::thisInstance()->indexReference
          ( QString((*it).second.info.c_str()) );
    
    if(idx && idx->getParentFolder() == this)
    {
      QString path = name()+"/"+idx->getID();
    
      if( idx->getUnreadMark() )
        decrementUnread();  
    
      indexCollection.remove( idx->getID() );
      
      paths += path;
    }
  }
  
  {
    // synch the uidJar
    for(UIDMap::iterator it=uidJar->rcvList.begin(); it!=uidJar->rcvList.end(); ++it)
    {
      IndexClass *idx = ObjectRequestBroker::thisInstance()->indexReference(QString((*it).second.info.c_str()));
      if( !idx )
      {
        uidJar->rcvList.erase(it);
      }
    }
  }
  
  uidJar->trashList.clear();
  uidJar->saveLists();
  
  compact();
  
  // notify the client
  ServerNotifier::thisInstance()->objectDeleted(paths);
  
  return true;
}

bool RemoteMailFolder::deleteMessage(const QString & indexURL)
{
  IndexClass * idx = ObjectRequestBroker::thisInstance()->indexReference( indexURL );
  if(!idx || idx->getParentFolder()!=this)
    return false;

  QString path=name()+"/"+idx->getID();

  if(idx->getUnreadMark())
    decrementUnread();

  indexCollection.remove(idx->getID());
  
  compact();

  // notify the client
  ServerNotifier::thisInstance()->objectDeleted(path);

  return true;
}

QString RemoteMailFolder::copyMessage(IndexClass *idx)
{
  if(idx)
  {
    MailFolder * mailFolder = idx->getParentFolder();
    if(!mailFolder)
      return QString::null;

    
    if( MailFolder::IMAP4 == mailFolder->getFolderProtocol() &&
        !idx->isFullMessage() )
    {
      QStringList param;
      
      IOSyncHandler::ref()->insertToDo(idx->url(), name(), WaitingOperations::CopyObject, param);
      IMAP4Handler::ref()->updateMessage(*idx);
    }
    else
    {
      IMAP4Handler::ref()->addMessage(name(), idx);
    }
  }
  
  // return null because the new mail will be created with createMessage()
  return QString::null;
}

QString RemoteMailFolder::moveMessage(IndexClass *idx)
{
  if(idx)
  {
    MailFolder * mailFolder = idx->getParentFolder();
    if(!mailFolder)
      return QString::null;

    QStringList param;
      
    if( MailFolder::IMAP4 == mailFolder->getFolderProtocol() &&
        !idx->isFullMessage() )
    {
      IOSyncHandler::ref()->insertToDo(idx->url(), name(), WaitingOperations::MoveObject, param);
      IMAP4Handler::ref()->updateMessage(*idx);
    }
    else
    {
      QString folderURL = mailFolder->name();
      
      IOSyncHandler::ref()->insertToDo(idx->url(), folderURL, WaitingOperations::DeleteObject, param);
      IMAP4Handler::ref()->addMessage(name(), idx);
    }
  }
  
  // return null because the new mail will be created with createMessage()
  return QString::null;
}

bool RemoteMailFolder::expunge(bool force)
{
  if( force || shouldExpunge() )
  {
    // TODO: Remove the indexes older then a date
    // or the oldest mails till the folder size is good.
    
    // call the expunge command
    
    compact();
    
    return true;
  }
  
  return false;
}

bool RemoteMailFolder::loadStatus()
{
  QString group = QString("/Aethera/IMAPStatus/")+name()+"/";

  status.messages=
    QSettingsExtra::readUnsignedLongNumEntry
    (Settings::instance(), group+"MESSAGES", 0);

  status.recent=
    QSettingsExtra::readUnsignedLongNumEntry
    (Settings::instance(), group+"RECENT", 0);

  status.unseen=
    QSettingsExtra::readUnsignedLongNumEntry
    (Settings::instance(), group+"UNSEEN", 0);

  status.uidNext=
    QSettingsExtra::readUnsignedLongNumEntry
    (Settings::instance(), group+"UIDNEXT", 0);

  status.uidValidity=
    QSettingsExtra::readUnsignedLongNumEntry
    (Settings::instance(), group+"UIDVALIDITY", 0);
  
  return true;
}

bool RemoteMailFolder::saveStatus()
{
  QString group = QString("/Aethera/IMAPStatus/")+name()+"/";

  QSettingsExtra::writeEntry
    (Settings::instance(), group+"MESSAGES", status.messages);

  QSettingsExtra::writeEntry
    (Settings::instance(), group+"RECENT", status.recent);

  QSettingsExtra::writeEntry
    (Settings::instance(), group+"UNSEEN", status.unseen);

  QSettingsExtra::writeEntry
    (Settings::instance(), group+"UIDNEXT", status.uidNext);

  QSettingsExtra::writeEntry
    (Settings::instance(), group+"UIDVALIDITY", status.uidValidity);
  
  return true;
}

bool RemoteMailFolder::removeStatus()
{
  QString group = QString("/Aethera/IMAPStatus/") + name();

  return Settings::instance()->removeGroup( group );
}

unsigned long RemoteMailFolder::messages() const
{
  return status.messages;
}

unsigned long RemoteMailFolder::recent() const
{
  return status.recent;
}

unsigned long RemoteMailFolder::unseen() const
{
  return status.unseen;
}

unsigned long RemoteMailFolder::uidNext() const
{
  return status.uidNext;
}

unsigned long RemoteMailFolder::uidValidity() const
{
  return status.uidValidity;
}

void RemoteMailFolder::setStatus( unsigned long uidValidity, unsigned long uidNext,
                                  unsigned long messages, unsigned long unseen,
                                  unsigned long recent)
{
  status.uidValidity=uidValidity;
  status.uidNext=uidNext;
  status.messages=messages;
  status.unseen=unseen;
  status.recent=recent;
}


void RemoteMailFolder::showStatus() const
{
  qDebug("IMAP Folder Status");
  qDebug("UID Validity=%lu", status.uidValidity);
  qDebug("UID Next=%lu", status.uidNext);
  qDebug("Messages=%lu", status.messages);
  qDebug("Unseen=%lu", status.unseen);
  qDebug("Recent=%lu", status.recent);
}

bool RemoteMailFolder::moveMessage(const QStringList & urls)
{
  IndexClass * idx = ObjectRequestBroker::thisInstance()->indexReference(urls[0]);
  MailFolder * mailFolder = idx->getParentFolder();
  if( !mailFolder )
    return false;
  
  QStringList param;
        
  for(QStringList::ConstIterator it=urls.begin(); it!=urls.end(); ++it)
  {
    idx = mailFolder->indexIDict()[ ObjectRequestBroker::thisInstance()->object(*it) ];
    if(idx)
    {
      if( MailFolder::IMAP4 == mailFolder->getFolderProtocol() &&
          !idx->isFullMessage() )
      {
        IOSyncHandler::ref()->insertToDo(idx->url(), name(), WaitingOperations::MoveObject, param);
        IMAP4Handler::ref()->updateMessage(*idx);
      }
      else
      {
        QString folderURL = mailFolder->name();
        
        IOSyncHandler::ref()->insertToDo(idx->url(), folderURL, WaitingOperations::DeleteObject, param);
        IMAP4Handler::ref()->addMessage(name(), idx);
      }
    }
  }
  
  return true;
}

bool RemoteMailFolder::copyMessage(const QStringList & urls)
{
  IndexClass * idx = ObjectRequestBroker::thisInstance()->indexReference(urls[0]);
  MailFolder * mailFolder = idx->getParentFolder();
  if( !mailFolder )
    return false;
  
  for(QStringList::ConstIterator it=urls.begin(); it!=urls.end(); ++it)
  {
    idx = mailFolder->indexIDict()[ ObjectRequestBroker::thisInstance()->object(*it) ];
    if(idx)
    {
      if( MailFolder::IMAP4 == mailFolder->getFolderProtocol() &&
          !idx->isFullMessage() )
      {
        QStringList param;
        
        IOSyncHandler::ref()->insertToDo(idx->url(), name(), WaitingOperations::CopyObject, param);
        IMAP4Handler::ref()->updateMessage(*idx);
      }
      else
      {
        IMAP4Handler::ref()->addMessage(name(), idx);
      }
    }
  }
  
  return true;
}

bool RemoteMailFolder::deleteMessage(const QStringList & urls)
{
  IndexClass * idx = 0L;
  
  for(QStringList::ConstIterator it=urls.begin(); it!=urls.end(); ++it)
  {
    idx = indexCollection[ ObjectRequestBroker::thisInstance()->object(*it) ];
    if( idx )
    {
      if( !IMAP4Handler::ref()->deleteMessage(*idx) )
        return false;
    }
  }
  
  return true;
}

