/*
    Copyright (C) 2001-2002 by theKompany.com <www.thekompany.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    By PLUGINS we mean the commercial plug-ins developed by theKompany.com
    for Aethera.

    In addition, as a special exception, theKompany.com gives permission
    to link the code of this program with PLUGINS (or with
    modified versions of PLUGINS that use the same license as PLUGINS),
    and distribute linked combinations including the two.  You must obey
    the GNU General Public License in all respects for all of the code used
    other than PLUGINS.  If you modify this file, you may extend this
    exception to your version of the file, but you are not obligated to do so.
    If you do not wish to do so, delete this exception statement from your
    version.

    This license grants you the ability to use PLUGINS with Aethera only
    and may not be used outside of Aethera.
    See also http://www.thekompany.com/products/license.txt for details.
*/
#ifndef __UIPART_H
#define __UIPART_H

#include <kore/interface.h>
#include <kore/metainfo.h>
#include <kore/componentref.h>
#include <kore/componentproxy.h>

// Includes for Qt classes :(
#include <qwidget.h>
#include <qptrlist.h>

// Forwards for Qt classes
class QMainWindow;
class QMenuBar;
class QStatusBar;
class QString;

#ifdef Q_WS_WIN
	#include <windows.h>
	#ifdef TINO_EXPORTS
		#define TINO_API __declspec(dllexport)
	#else
		#define TINO_API __declspec(dllimport)
	#endif
#else
  #define TINO_API
#endif

#define PART_SERVICE "/TINO/Part"
#define PART_SERVICE_DESC "The part service from TINO"

#define TOOLBARS    "toolbars"
#define MENUBARS    "menus"
#define WIDGET      "widget"

/**
 * class UIPart - you're supposed to extend it, implement the abstract functions, etc...
 * An instance of this is fed to the part manager (see below), who handles the UI merging,
 * perhaps moving the widget around, etc...
 */

// Forward declaration, you don't really want to know about this :)
class UIPartPrivate;

class TINO_API UIPart: public kore::ComponentRef, public kore::Interface
{
  public:
    UIPart(void);
    virtual ~UIPart();
    /*
     * Informative function - its aim is to tell the part manager what it
     * should care about.
     *
     * Capabilities queried for:
     * "widget"
     * "toolbars"
     * "menus"
     * The above are self-explaining
     * "clickactivate"
     * "dblclickactivate"
     * NOTE that the mouse event that activated the part is not transmitted to the
     * part's widget, as opposed to KParts which lets it go, and that the part
     * won't be activated by mouse clicks if you don't specify either clickactivate
     * or dblclickactivate
     * "activenotify"
     * If you answer true to this one, the activated() and deactivated() abstract methods
     * defined below will be called on activation/deactivation. Don't forget to reimplement
     * them...
     */
    virtual bool hasCapability(const QString& capname);
    /*
     * If there is a widget, return it to the part manager here. If hasWidget (above)
     * returns false, this function will never be called.
     */
    virtual QWidget *getWidget(const QString&) = 0;
    /*
     * This function is called by the manager to notify your component to create its main
     * widget. It expects the widget to be always available afterwards.
     * Creating your (embedded) widget here is not compulsory, but if you have your reasons
     * not to do it in the constructor, here you go...
     * Note that you *need* to reimplement it if you answer true to hasCapability("widget"),
     * even if you don't use it - it's a feature of C++ :)
     */
    virtual void createWidget(const QString&) = 0;
    /*
     * Called by the manager in addPart - don't bother about it
     */
    void linkWidget(const QString&);
    /*
     * Primitive menu/toolbar integration
     * The show* functions expect you to make your toolbars/menus available and provides a
     * QMainWindow/QMenuBar for you to attach them to.
     * You should destroy/hide your items when the hide* functions are called, your choice
     * if you keep them for the lifetime of the object or create/destroy them on each call.
     */
    virtual void showToolbars(QMainWindow *) = 0;
    virtual void hideToolbars(QMainWindow *) = 0;
    virtual void showMenus(QMenuBar *) = 0;
    virtual void hideMenus(QMenuBar *) = 0;
    /*
     * These two functions are called on part activation/deactivation (on click, double click)
     * by default they just set the active flag, reimplement if you need.
     */
    virtual void activated(void);
    virtual void deactivated(void);
    /*
     * Call these functions to emit signals to the PartManager
     * They use the UIPartPrivate to do the actual signal job, we don't want
     * Qt derivation here
     * activated/deactivated are automagical, you don't actually need to call them,
     * but I left them here "just in case"
     */
    virtual void emitActivated(void);
    virtual void emitDeactivated(void);
    virtual void emitWantUIReload(void);
    virtual void statusMessage(QString& message);
    virtual void statusMessage(QString& message, int ms);
    /*
     * Status function
     */
    bool isActive();
    /*
     * Could be made private, but then UIPartManager needs to be a friend...
     * and friend classes are pure evil(tm)
     */
    UIPartPrivate *getQTHandler(void);
    
  protected:
    /*
     * Reimplement this if you need your own signal handler (UIPartPrivate)
     */
    virtual UIPartPrivate *createQTHandler(void);
    
  private:
    UIPartPrivate *qthandler;
    bool partActive;
};


/*
 * The goal of this little class is to be revealed into a future version of this API
 * Don't worry, it's for "OLE"
class ContainerWidget: public QWidget
{ Q_OBJECT
  public:
    ContainerWidget(QWidget *parent=0, const char *name=0, WFlags f=0); // Adauga parametrii de la QWidget
    virtual ~ContainerWidget();
};
*/

class UIPartFactory: public kore::ComponentRef, kore::Interface
{
  public:
    UIPartFactory(void);
    virtual ~UIPartFactory() = 0;
    virtual UIPart *createUIPart(void) = 0;
};

#endif

