/*
    Copyright (C) 2001-2002 by theKompany.com <www.thekompany.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    By PLUGINS we mean the commercial plug-ins developed by theKompany.com
    for Aethera.

    In addition, as a special exception, theKompany.com gives permission
    to link the code of this program with PLUGINS (or with
    modified versions of PLUGINS that use the same license as PLUGINS),
    and distribute linked combinations including the two.  You must obey
    the GNU General Public License in all respects for all of the code used
    other than PLUGINS.  If you modify this file, you may extend this
    exception to your version of the file, but you are not obligated to do so.
    If you do not wish to do so, delete this exception statement from your
    version.

    This license grants you the ability to use PLUGINS with Aethera only
    and may not be used outside of Aethera.
    See also http://www.thekompany.com/products/license.txt for details.
*/
#include "tktabbar.h"

#include <qapplication.h>
#include <qbitmap.h>
#include <qpainter.h>
#include <qdrawutil.h>
#include <qregion.h>
#include <qaccel.h>
#include <qstyle.h>

struct TKTab
{
  QString caption;
  QIconSet *icon;
  int id;
  int width;
  bool enabled;
};

struct TKTabBarPrivate
{
  QPointArray pbreak;
  QPointArray pleft;
  QPointArray pright;
  QPointArray pclose;
  QPtrList<TKTab> tabs;
};

TKTabBar::TKTabBar(QWidget *parent, const char *name)
: QWidget(parent, name, WResizeNoErase | WRepaintNoErase)
{
  d = new TKTabBarPrivate;
  updatePixmap();

  buffer.resize(1, 1);

  accel = new QAccel(this);
  connect(accel, SIGNAL(activated(int)), SLOT(setCurrentTab(int)));

  connect(&breakTimer, SIGNAL(timeout()), SLOT(breakTimeout()));

  breakLeft = false;
  breakRight = false;
  leftButtonPressed = false;
  rightButtonPressed = false;
  closeButtonPressed = false;
  mousePressed = false;

  d->tabs.setAutoDelete(true);
  current = -1;
  first = -1;
  tabPressed = -1;

  vstyle = ShowIcon;
  showDragLine = false;
  easyStyle = false;
  selectedBold = false;
  enableClose = false;
  showIcon = true;
  controlsWidth = 38;

  setFocusPolicy(TabFocus);
  setBackgroundMode(NoBackground);

  setTabPosition(Top);
}

TKTabBar::~TKTabBar()
{
  TKTab *t = d->tabs.first();
  while (t) {
    accel->removeItem(t->id);
    delete t->icon;
    d->tabs.removeRef(t);
    t = d->tabs.first();
  }
  current = -1;
  first = -1;
  delete d;
}

int TKTabBar::count() const
{
  return d->tabs.count();
}

void TKTabBar::setViewStyle(int style)
{
  vstyle = style;

  showDragLine = vstyle & ShowDragLines;
  easyStyle = vstyle & EasyStyle;
  selectedBold = vstyle & CurrentBold;
  enableClose = vstyle & ShowCloseButton;
  showIcon = vstyle & ShowIcon;

  controlsWidth = 38 + (enableClose ? 14 : 0);
  buffer.resize(QMAX(buffer.width(), controlsWidth), buffer.height());

  recalTabsWidth();
  updateBreaks();
  updateGeometry();
  repaint(false);
}

int TKTabBar::viewStyle()
{
  return vstyle;
}

void TKTabBar::recalTabsWidth()
{
  int maxw = 0;
  for (TKTab *t = d->tabs.first(); t; t = d->tabs.next()) {
    t->width = tabWidth(t);
    maxw = QMAX(maxw, t->width);
  }
  buffer.resize(QMAX(buffer.width(), maxw + 4), buffer.height());
}

TKTab *TKTabBar::tab(int id) const
{
  TKTabBar *self = (TKTabBar *)this;
  for (TKTab *t = self->d->tabs.first(); t; t = self->d->tabs.next())
    if (t->id == id)
      return t;

  return 0;
}

void TKTabBar::polish()
{
  QWidget::polish();
  if (first == -1)
    first = firstTab();
  if (current == -1)
    setCurrentTab(firstTab(true));
}

int TKTabBar::insertTab(const QString &caption, QIconSet *icon, int idAfter, int id)
{
  while (id == -1 || tab(id))
    id++;

  TKTab *t = new TKTab;
  t->caption = caption;
  t->icon = icon;
  t->id = id;
  t->width = tabWidth(t);
  t->enabled = true;

  int a = QAccel::shortcutKey(caption);
  if (a)
    accel->insertItem(a, id);

  buffer.resize(QMAX(buffer.width(), t->width + 4), buffer.height());

  (void)tab(idAfter); // set current list item to idAfter, or -1 if idAfter not exist
  d->tabs.insert(d->tabs.at() + 1, t);

  updateBreaks();
  return id;
}

int TKTabBar::insertTab(const QString &caption, int idAfter, int id)
{
  return insertTab(caption, QIconSet(), idAfter, id);
}

int TKTabBar::insertTab(const QString &caption, const QIconSet &icon, int idAfter, int id)
{
  return insertTab(caption, new QIconSet(icon), idAfter, id);
}

int TKTabBar::addTab(const QString &caption, const QIconSet &icon, int id)
{
  return insertTab(caption, new QIconSet(icon), lastTab(), id);
}

int TKTabBar::addTab(const QString &caption, int id)
{
  return insertTab(caption, QIconSet(), lastTab(), id);
}

void TKTabBar::setTabPosition(TabPosition pos)
{
  a = pos;
  updateGeometry();
  repaint(false);
}

int TKTabBar::rw()
{
  return (a == Left || a == Right) ? height() : width();
}

int TKTabBar::rh()
{
  return (a == Left || a == Right) ? width() : height();
}

QSizePolicy TKTabBar::sizePolicy() const
{
  return (a == Left || a == Right) ?
          QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred) :
          QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
}

QSize TKTabBar::sizeHint() const
{
  int w = 5 + 20 + 3 + controlsWidth;
  int h = QMAX(showIcon ? 14 : 0, fontMetrics().height()) + 10 + 2;

  return ((a == Left || a == Right) ? QSize(h, w) : QSize(w, h)).expandedTo( QApplication::globalStrut());
}

void TKTabBar::paintEvent(QPaintEvent *)
{
  const QColorGroup &g = colorGroup();
  QColor bg = easyStyle ? g.midlight() : g.background();

  QWMatrix tm;
  QColor cl;
  QColor cr;
  QColor ct;
  QColor cb;
  switch (a) {
    case Top:
      cl = g.light();
      cr = g.dark();
      ct = g.light();
      cb = g.light();
      break;
    case Bottom:
      tm.scale(1, -1);
      cl = g.light();
      cr = g.dark();
      ct = g.dark();
      cb = g.dark();
      break;
    case Left:
      tm.rotate(90);
      tm.scale(1, -1);
      cl = g.light();
      cr = g.dark();
      ct = g.light();
      cb = g.light();
      break;
    case Right:
      tm.rotate(90);
      cl = g.light();
      cr = g.dark();
      ct = g.dark();
      cb = g.dark();
      break;
  }
  QPainter paint(this, this);

  QPixmap temp; // bug?
  QPainter pb(&buffer, this);
/********************************************************************************************************/
  // draw left break or small line before first tab if it's tab not current
  if (breakLeft) {                          // left break, width = 5 ( + 2 )
    int w = 5 + (current != first ? 2 : 0); // if first visible is active
    pb.fillRect(0, 0, w, buffer.height(), bg);

    QPointArray br = d->pbreak.copy();
    pb.setPen(g.light());
    br.translate(0, 6);
    for (int y = 6; y < buffer.height(); y += 12) {
      pb.drawPoints(br);
      br.translate(0, 12);
    }

    if (!easyStyle) {
      pb.setPen(ct);
      pb.drawLine(1, 5, w - 1, 5); // top
      pb.setPen(cr);
      pb.drawLine(w - 1, 5, w - 1, buffer.height() - 2); // right
    }
    pb.setPen(cb);
    pb.drawLine(0, buffer.height() - 1, w - 1, buffer.height() - 1); // bottom

    temp = buffer; // bug?
    switch (a) {
      case Top:
      case Bottom:
        paint.drawPixmap(0, 0, temp.xForm(tm), 0, 0, w, height());
        break;
      case Left:
      case Right:
        paint.drawPixmap(0, 0, temp.xForm(tm), 0, 0, buffer.height(), w);
        break;
    }
  } else {
    int w = startPos();
    pb.fillRect(0, 0, w, buffer.height(), bg);
    pb.setPen(cb);
    pb.drawLine(0, buffer.height() - 1, w - 1, buffer.height() - 1); // bottom line from start to beginning first tab

    temp = buffer; // bug?
    switch (a) {
      case Top:
      case Bottom:
        paint.drawPixmap(0, 0, temp.xForm(tm), 0, 0, w, height());
        break;
      case Left:
      case Right:
        paint.drawPixmap(0, 0, temp.xForm(tm), 0, 0, buffer.height(), w);
        break;
    }
  }
/********************************************************************************************************/
  // draw tabs
  bool v = false;
  bool lastIsCurrent = false;
  bool breakR = false;
  int x = startPos();

  QWMatrix rm;
  rm.rotate(90);
  QWMatrix lm;
  lm.rotate(-90);

  int beforeCurrent = prevTab(current);
  int afterCurrent = tabRect(current).isValid() ? nextTab(current) : -1;

  for (TKTab *t = d->tabs.first(); t && !breakR; t = d->tabs.next()) {
    if (!v && t->id == first)
      v = true;
    if (v) {
      int fx = x + (t->id == current ? -2 : 0) + (t->id == afterCurrent ? 2 : 0);
      int dw = t->width + (t->id == current ? 4 : 0) - (t->id == beforeCurrent || t->id == afterCurrent ? 2 : 0);

      pb.fillRect(0, 0, buffer.width(), buffer.height(), bg);
      drawTab(pb, t->enabled ? g : palette().disabled(), t);
      temp = buffer; // bug?
      switch (a) {
        case Bottom:
        case Top:
          paint.drawPixmap(fx, 0, temp, t->id == afterCurrent ? 2 : 0, 0, QMIN(dw, width() - controlsWidth - 5 - fx), height());
          break;
        case Left:
          paint.drawPixmap(0, fx, temp.xForm(lm), 0, buffer.width() - dw - (t->id == beforeCurrent ? 2 : 0), width(), QMIN(dw, height() - controlsWidth - 5 - fx));
          break;
        case Right:
          paint.drawPixmap(0, fx, temp.xForm(rm), 0, t->id == afterCurrent ? 2 : 0, width(), QMIN(dw, height() - controlsWidth - 5 - fx));
          break;
        default:
          break;
      }
      x += t->width;
      breakR = x > (rw() - controlsWidth - 5 - (t->id == current ? 2 : 0));
      lastIsCurrent = t->id == current;
    }
  }
/********************************************************************************************************/
  if (x < rw() - controlsWidth - 5) {
    // fill unused space and
    // draw bottom line from end of last tab to starting right break
    x += lastIsCurrent ? 2 : 0;
    switch (a) {
      case Top:
        paint.fillRect(x, 0, width() - controlsWidth - 5 - x, height(), bg);
        paint.setPen(g.light());
        paint.drawLine(x, height() - 1, width() - controlsWidth - 5, height() - 1);
        break;
      case Bottom:
        paint.fillRect(x, 0, width() - controlsWidth - 5 - x, height(), bg);
        paint.setPen(g.dark());
        paint.drawLine(x, 0, width() - controlsWidth - 5, 0);
        break;
      case Left:
        paint.fillRect(0, x, width(), height() - controlsWidth - 5 - x, bg);
        paint.setPen(g.light());
        paint.drawLine(width() - 1, x, width() - 1, height() - controlsWidth - 5);
        break;
      case Right:
        paint.fillRect(0, x, width(), height() - controlsWidth - 5 - x, bg);
        paint.setPen(g.dark());
        paint.drawLine(0, x, 0, height() - controlsWidth - 5);
        break;
    }

    paint.setPen(white); // bug?
  }
/********************************************************************************************************/
  // draw right break or small line if rightBreak not visible
  bool breakRightT = (tabRect(lastTab()).right() >= rw() - controlsWidth - 5);
  if (breakRightT) {  // right break, width = 5
    if (lastIsCurrent) {
      pb.fillRect(0, 0, 5, 2, bg);
      pb.fillRect(0, 3, 5, buffer.height(), g.background());

      if (!easyStyle) {
        pb.setPen(cb);
        pb.drawLine(3, buffer.height() - 1, 4, buffer.height() - 1); // bottom
      }
      QPointArray br = d->pbreak.copy();
      pb.setPen(g.dark());
      br.translate(2, 3);
      for (int y = 6; y < buffer.height(); y += 12) {
        pb.drawPoints(br);
        br.translate(0, 12);
      }

      pb.setPen(ct);
      pb.drawLine(0, 3, 2, 3); // top
    } else {
      pb.fillRect(0, 0, 5, buffer.height(), bg);

      QPointArray br = d->pbreak.copy();
      pb.setPen(g.dark());
      br.translate(2, 5);
      for (int y = 6; y < buffer.height(); y += 12) {
        pb.drawPoints(br);
        br.translate(0, 12);
      }
      if (!easyStyle) {
        pb.setPen(ct);
        pb.drawLine(0, 5, 2, 5); // top
        pb.setPen(cb);
        pb.drawLine(0, buffer.height() - 1, 4, buffer.height() - 1); // bottom
      }
    }
  } else {
    pb.fillRect(0, 0, 5, buffer.height(), bg);
    pb.setPen(cb);
    pb.drawLine(0, buffer.height() - 1, 5, buffer.height() - 1); // bottom line if rightBreak not visible
  }
  temp = buffer; // bug?
  switch (a) {
    case Top:
    case Bottom:
      paint.drawPixmap(width() - controlsWidth - 5, 0, temp.xForm(tm), 0, 0, 5, height());
      break;
    case Left:
    case Right:
      paint.drawPixmap(0, height() - controlsWidth - 5, temp.xForm(tm), 0, 0, buffer.height(), 5);
      break;
  }
/********************************************************************************************************/
  // panel for left-right buttons, width = controlsWidth, or line
  pb.fillRect(0, 0, controlsWidth, buffer.height(), bg);
  if (breakLeft || breakRight || enableClose) {
    if (!easyStyle) {
      pb.setPen(cr);
      pb.drawLine(controlsWidth - 1, 3, controlsWidth - 1, buffer.height() - 1); // right
      pb.setPen(ct);
      pb.drawLine(controlsWidth - 2, 3, 5, 3);
      pb.drawLine(5, 3, 3, 5);
      pb.setPen(cl);
      pb.drawLine(3, 5, 3, buffer.height() - 1);
      pb.setPen(cb);
      pb.drawLine(3, buffer.height() - 1, 0, buffer.height() - 1);
    } else {
      pb.setPen(cb);
      pb.drawLine(0, buffer.height() - 1, buffer.width() - 1, buffer.height() - 1);
    }
    // draw arrows
    int at = (buffer.height() - 11) / 2 + 2;
    int x1 = 10;
    int x2 = x1 + 14;
    int x3 = x2 + 14;

    pb.setPen(breakLeft && isEnabled() ? g.foreground() : g.mid());
    int down = leftButtonPressed ? 1 : 0;

    d->pleft.translate(x1 + down, at + down);
    pb.drawPoints(d->pleft);
    d->pleft.translate(-(x1 + down), -(at + down));

    pb.setPen(breakRight && isEnabled() ? g.foreground() : g.mid());
    down = rightButtonPressed ? 1 : 0;

    d->pright.translate(x2 + down, at + down);
    pb.drawPoints(d->pright);
    d->pright.translate(-(x2 + down), -(at + down));

    if (enableClose) {
      pb.setPen(firstTab() != -1 && isEnabled() ? g.foreground() : g.mid());
      down = closeButtonPressed ? 1 : 0;
      d->pclose.translate(x3 + down, at + down);
      pb.drawPoints(d->pclose);
      d->pclose.translate(-(x3 + down), -(at + down));
    }
  } else {
    pb.setPen(cb);
    pb.drawLine(0, buffer.height() - 1, controlsWidth - 1, buffer.height() - 1); // bottom line from buttonsPanel to end
  }
  temp = buffer; // bug?
  switch (a) {
    case Top:
    case Bottom:
//      paint.drawPixmap(width() - controlsWidth, 0, buffer, 0, 0, controlsWidth, height());
      paint.drawPixmap(width() - controlsWidth, 0, temp.xForm(tm), 0, 0, controlsWidth, height());
      break;
    case Left:
    case Right:
      paint.drawPixmap(0, height() - controlsWidth, temp.xForm(tm), 0, 0, buffer.height(), controlsWidth);
      break;
  }
/********************************************************************************************************/
  pb.end();
  paint.end();
}

void TKTabBar::drawTab(QPainter &paint, const QColorGroup &g, TKTab *t)
{
  bool enabled = t->enabled && isEnabled();
  bool active = t->id == current;

  int x = 0;
  int y = active ? 3 : 5;
  int w = t->width + (active ? 4 : 0);
/**************************************************************************/
  QWMatrix m;
  QColor cl;
  QColor cr;
  QColor ct;
  QColor cb;
  switch (a) {
    case Top:
      cl = g.light();
      cr = g.dark();
      ct = g.light();
      cb = g.light();
      break;
    case Bottom:
      m.scale(1, -1);
      m.translate(0, 1 - height());
      cl = g.light();
      cr = g.dark();
      ct = g.dark();
      cb = g.dark();
      break;
    case Left:
      cl = g.dark();
      cr = g.light();
      ct = g.light();
      cb = g.light();
      break;
    case Right:
      cl = g.light();
      cr = g.dark();
      ct = g.dark();
      cb = g.dark();
      break;
  }
  paint.save();
  paint.setWorldMatrix(m);

  if (!active && easyStyle)
    paint.fillRect(0, y, w, buffer.height(), g.midlight());
  else
    paint.fillRect(0, y, w, buffer.height(), g.background());

  if (!active && easyStyle) {
    paint.setPen(g.midlight().dark(120));
    if (a == Left)
      paint.drawLine(0, y, 0, buffer.height() - 2); // separator
    else
      paint.drawLine(w - 1, y, w - 1, buffer.height() - 2); // separator
    x += 3;
  } else {
    paint.setPen(ct);
    paint.drawLine(w - 1, y, x, y); // top

    paint.setPen(cl);
    paint.drawLine(x, y, x, buffer.height() - 1); // left

    paint.setPen(cr);
    paint.drawLine(w - 1, buffer.height() - 1, w - 1, y); // right
    x += 3;

    if (showDragLine) {
      x += 3;
      qDrawShadeLine(&paint, x, y + 3, x, buffer.height() - 3, g, false, 1, 0);
      x += 3;
      qDrawShadeLine(&paint, x, y + 3, x, buffer.height() - 3, g, false, 1, 0);
      x += 3;
    }
  }

  if (!active) {
    paint.setPen(cb);
    paint.drawLine(0, buffer.height() - 1, w - 1, buffer.height() - 1); // bottom
  }

  paint.restore();
/**************************************************************************/
  m.reset();
  switch (a) {
    case Bottom:
      m.translate(0, -3);
      break;
    default:
      break;
  }
  paint.save();
  paint.setWorldMatrix(m);

  if (showIcon && !t->icon->isNull()) {
    x += 2;
    QIconSet::Mode mode = enabled ? (active ? QIconSet::Active : QIconSet::Normal) : QIconSet::Disabled;
    QPixmap p(t->icon->pixmap(QIconSet::Small, mode));
    paint.drawPixmap(x, (buffer.height() - p.height() + 4) / 2, p);
    x += p.width();
  }
  x += 3;
/**************************************************************************/
  QColorGroup mg(g);
  QFont f(font());
  if (active) {
    if (selectedBold)
      f.setBold(true);
    paint.setFont(f);
  } else {
    if (mg.color(QColorGroup::Foreground)!=QColor(0,0,200))
  mg.setColor(QColorGroup::Foreground, g.dark().dark());
  }
  QFontMetrics fm(f);
  QRect tr(x, y, fm.width(clearString(t->caption)) + 1, buffer.height() - y - 5);

#if QT_VERSION >= 300
  style().drawItem(&paint, tr, AlignLeft | AlignBottom | ShowPrefix, mg, enabled, 0, t->caption);
#else
  style().drawItem(&paint, tr.x(), tr.y(), tr.width(), tr.height(), AlignLeft | AlignBottom | ShowPrefix, mg, enabled, 0, t->caption);
#endif

  if (enabled && active && hasFocus()) {
    paint.setPen(g.mid());
    paint.drawRect(x - 2, y + 3, tr.width() + 4, buffer.height() - y - 5);
  }
  paint.restore();
}
/**************************************************************************/
int TKTabBar::startPos()
{
  return (d->tabs.first() ? 2 : 0) + ((d->tabs.first() && d->tabs.first()->id != first ) ? 5 : (easyStyle ? 5 : 0));
}

QRect TKTabBar::tabRect(int id)
{
  int x = startPos();
  bool v = false;
  for (TKTab *t = d->tabs.first(); t; t = d->tabs.next()) {

    if (!v)
      if (t->id == first)
        v = true;
      else
        if (t->id == id)
          break;

    if (!v)
      continue;

    if (t->id != id)
      x += t->width;
    else
      return QRect(x + (current == id ? -2 : 0), 0, t->width + (current == id ? 4 : 0), rh());
  }
  return QRect();
}

QString TKTabBar::clearString(const QString &s)
{
  // remove shortcutKey for right calculate string width
  QString c = s;
  int i = -1;
  while (true) {
    i = c.find('&', ++i);
    if (i == -1)
      break;

    if (i != -1 && i != (int)c.length() - 1)
      c.remove(i, 1);
  }
  return c;
}

int TKTabBar::tabWidth(TKTab *t)
{
  QFont f(font());
  int w = 0;
  if (t->id == current) {
    if (selectedBold)
      f.setBold(true);
    w = showDragLine ? 12 : 4;
  } else {
    w = easyStyle ? 4 : 12;
  }
  QFontMetrics fm(f);
  w += fm.width(clearString(t->caption));

  if (showIcon && !t->icon->isNull())
    w += 16 + 3;

  w += 8;

  return w;
}

void TKTabBar::checkButtons(const QPoint &pos)
{
  QPoint rp = (a == Left || a == Right) ? QPoint(pos.y(), pos.x()) : QPoint(pos.x(), pos.y());
  int buttonY = (rh() - 11) / 2 + 2;
  QRect buttonRect(rw() - controlsWidth + 10, buttonY, 9, 11);
  leftButtonPressed = breakLeft && buttonRect.contains(rp);
  buttonRect.moveBy(14, 0);
  rightButtonPressed = breakRight && buttonRect.contains(rp);
  buttonRect.moveBy(14, 0);
  closeButtonPressed = buttonRect.contains(rp);
}

int TKTabBar::tabByPos(const QPoint &pos)
{
  QPoint rp = (a == Left || a == Right) ? QPoint(pos.y(), pos.x()) : QPoint(pos.x(), pos.y());
  if (rp.x() < startPos())
    return prevTab(first);

  if (rp.x() < rw() - controlsWidth)
    for (TKTab *t = d->tabs.first(); t; t = d->tabs.next())
      if (tabRect(t->id).contains(rp))
        return t->id;

  return -1;
}

void TKTabBar::mousePressEvent(QMouseEvent *ev)
{
  tabPressed = tabByPos(ev->pos());
  if (tabPressed != -1 && !tab(tabPressed)->enabled)
    tabPressed = -1;

  if (ev->button() == Qt::RightButton) {
    if (tabPressed != -1)
      setCurrentTab(tabPressed);
    return;
  }

  outside = false;
  checkButtons(ev->pos());

  if (leftButtonPressed || rightButtonPressed) {
    repaint(false);
    breakTimer.start(250);
  }
  if (closeButtonPressed)
    repaint(false);

  setCurrentTabInternal(tabPressed);
}

void TKTabBar::mouseReleaseEvent(QMouseEvent *ev)
{
  if (ev->button() == Qt::RightButton) {
    emit tabRightButtonPressed(tabPressed, ev->globalPos());
    tabPressed = -1;
    return;
  }

  breakTimer.stop();
  if (outside) {
    emit finishTabDrag(ev);
    return;
  }
  if (leftButtonPressed) {
    leftButtonPressed = false;
    repaint(false);
    scrollRight();
  }
  if (rightButtonPressed) {
    rightButtonPressed = false;
    repaint(false);
    scrollLeft();
  }
  if (closeButtonPressed) {
    closeButtonPressed = false;
    repaint(false);
    emit closeTab(current);
    return;
  }

  if (tabPressed != -1)
    setCurrentTab(tabPressed);
  tabPressed = -1;
}

void TKTabBar::mouseMoveEvent(QMouseEvent *ev)
{
  if (ev->state() == Qt::RightButton)
    return;

  if (!rect().contains(ev->pos())) {
    if (!outside) {
      outside = true;
      emit startTabDrag(tabPressed, ev);
    } else {
      emit continueTabDrag(ev);
    }
    return;
  }

  if (outside) {
    outside = false;
    emit cancelTabDrag(ev);
  }

  bool oldLeft = leftButtonPressed;
  bool oldRight = rightButtonPressed;
  bool oldClose = closeButtonPressed;
  checkButtons(ev->pos());

  if (leftButtonPressed || rightButtonPressed)
    breakTimer.start(500);
  else
    breakTimer.stop();

  if ((oldLeft != leftButtonPressed) || (oldRight != rightButtonPressed) || (oldClose != closeButtonPressed))
    repaint(false);

  if (tabPressed != -1) {
    int currentTab = tabByPos(ev->pos());
    if (currentTab != tabPressed && currentTab != -1) {
      int oldFirst = first;
      if (tabPressed == first)
        first = nextTab(first);
      else
        if (currentTab == first)
          first = tabPressed;

      TKTab *t = tab(tabPressed);
      int id1 = d->tabs.at();
      tab(currentTab);
      int id2 = d->tabs.at();
      d->tabs.take(id1);
      d->tabs.insert(id2, t);
      if (tabByPos(ev->pos()) != tabPressed) { // revert changes
        tab(tabPressed);
        d->tabs.take();
        d->tabs.insert(id1, t);
        first = oldFirst;
      } else {
        repaint(false);
        emit changeTabsOrder(id1, id2);
      }
    }
  }
}

void TKTabBar::resizeEvent(QResizeEvent *ev)
{
  buffer.resize(buffer.width(), rh());

  if (d->tabs.first() && d->tabs.first()->id != first) {
    int w = 0;
    for (TKTab *t = d->tabs.first(); t; t = d->tabs.next())
      w += t->width;

    if (w < rw() - controlsWidth - 5) {
      first = d->tabs.first()->id;
    }
  }
  updateBreaks();
  QWidget::resizeEvent(ev);
}

void TKTabBar::scrollLeft()
{
  if (!breakRight)
    return;

  first = nextTab(first);
  updateBreaks();
  repaint(false);
}

void TKTabBar::scrollRight()
{
  if (!breakLeft)
    return;

  first = prevTab(first);
  updateBreaks();

  repaint(false);
}

void TKTabBar::updateBreaks()
{
  if (!d->tabs.first()) {
    breakLeft = false;
    breakRight = false;
  } else {
    breakLeft = d->tabs.first()->id != first;
    breakRight = (tabRect(lastTab()).right() >= rw() - controlsWidth - 5) && first != lastTab();
  }
  repaint(false);
}

int TKTabBar::firstTab(bool enabled)
{
  for (TKTab *t = d->tabs.first(); t; t = d->tabs.next())
    if (!enabled || (enabled && t->enabled))
      return t->id;
  return -1;
}

int TKTabBar::lastTab(bool enabled)
{
  for (TKTab *t = d->tabs.last(); t; t = d->tabs.prev())
    if (!enabled || (enabled && t->enabled))
      return t->id;
  return -1;
}

int TKTabBar::prevTab(int id, bool enabled)
{
  bool f = false;
  for (TKTab *t = d->tabs.last(); t; t = d->tabs.prev()) {
    if (f && (!enabled || (enabled && t->enabled)))
      return t->id;
    f = f || (t->id == id);
  }
  return -1;
}

int TKTabBar::nextTab(int id, bool enabled)
{
  bool f = false;
  for (TKTab *t = d->tabs.first(); t; t = d->tabs.next()) {
    if (f && (!enabled || (enabled && t->enabled)))
      return t->id;
    f = f || (t->id == id);
  }
  return -1;
}

void TKTabBar::setCurrentTab(int id)
{
  if (!setCurrentTabInternal(id))
    return;

  makeTabVisible(id);
  emit currentChanged(id);
}

bool TKTabBar::setCurrentTabInternal(int id)
{
  if (!tab(id))
    return false;

  int old = current;
  current = id;
  if (tab(old)) {
    tab(old)->width = tabWidth(tab(old));
    buffer.resize(QMAX(buffer.width(), tab(old)->width + 4), buffer.height());
  }
  tab(current)->width = tabWidth(tab(current));
  buffer.resize(QMAX(buffer.width(), tab(current)->width + 4), buffer.height());

  if (d->tabs.first() && d->tabs.first()->id != first) {
    int w = 0;
    for (TKTab *t = d->tabs.first(); t; t = d->tabs.next())
      w += t->width;

    if (w < rw() - controlsWidth - 5) {
      first = d->tabs.first()->id;
    }
  }

  updateBreaks();
  repaint(false);
  return true;
}

void TKTabBar::makeTabVisible(int id)
{
  bool v = false;
  for (TKTab *t = d->tabs.first(); t; t = d->tabs.next())
    if (v) {
      QRect r = tabRect(t->id);
      if (t->id != first && (r.right() > rw() - controlsWidth - 5))
        first = nextTab(first);
      if (t->id == id)
        break;
    } else {
      if (t->id == id) {
        first = id;
        break;
      }
      v = t->id == first;
    }

  updateBreaks();
}

void TKTabBar::breakTimeout()
{
  if (leftButtonPressed) {
    repaint(false);
    scrollRight();
  }
  if (rightButtonPressed) {
    repaint(false);
    scrollLeft();
  }
}

void TKTabBar::keyPressEvent(QKeyEvent *ev)
{
  int c = -1;
  switch (ev->key()) {
    case Key_Left:
    case Key_Up:
      c = prevTab(current, true);
      break;
    case Key_Right:
    case Key_Down:
      c = nextTab(current, true);
      break;
    case Key_Home:
      c = firstTab(true);
      break;
    case Key_End:
      c = lastTab(true);
      break;
    default:
      ev->ignore();
      return;
  }
  if (c != current && c != -1)
    setCurrentTab(c);
}

void TKTabBar::clear()
{
  TKTab *t = d->tabs.first();
  while (t) {
    accel->removeItem(t->id);
    delete t->icon;
    d->tabs.removeRef(t);
    t = d->tabs.first();
  }
  current = -1;
  first = -1;
  updateBreaks();
  emit currentChanged(-1);
  repaint(false);
}

void TKTabBar::removeTab(int id)
{
  TKTab *t = tab(id);
  if (t) {
    int c = -1;
    bool wasCurrent = t->id == current;
    if (wasCurrent) {
      c = prevTab(t->id, true);
      if (c == -1)
        c = nextTab(t->id, true);
    }
    accel->removeItem(t->id);
    delete t->icon;
    d->tabs.removeRef(t);

    if (wasCurrent) {
      if (c == -1) {
        current = -1;
        first = -1;
        updateBreaks();
        emit currentChanged(-1);
      } else
        setCurrentTab(c);
    }
    repaint(false);
  }
}

void TKTabBar::setTabTitle(int id, const QString &title)
{
  TKTab *t = tab(id);
  if (t) {
    t->caption = title;
    t->width = tabWidth(t);

    accel->removeItem(t->id);
    int a = QAccel::shortcutKey(t->caption);
    if (a)
      accel->insertItem(a, t->id);

    buffer.resize(QMAX(buffer.width(), t->width + 4), buffer.height());
    updateBreaks();
    repaint(false);
  }
}

void TKTabBar::setTabIcon(int id, const QIconSet &icon)
{
  TKTab *t = tab(id);
  if (t) {
    delete t->icon;
    t->icon = new QIconSet(icon);
    t->width = tabWidth(t);

    buffer.resize(QMAX(buffer.width(), t->width + 4), buffer.height());
    updateBreaks();
    repaint(false);
  }
}

void TKTabBar::setTabEnabled(int id, bool enabled)
{
  TKTab *t = tab(id);
  if (t) {
    t->enabled = enabled;
    repaint(false);
  }
}

QString TKTabBar::tabTitle(int id) const
{
  TKTab *t = tab(id);
  return t ? t->caption : QString::null;
}

QIconSet *TKTabBar::tabIcon(int id) const
{
  TKTab *t = tab(id);
  return t ? t->icon : 0;
}

bool TKTabBar::isTabEnabled(int id) const
{
  TKTab *t = tab(id);
  return t ? t->enabled : false;
}

void TKTabBar::fontChange(const QFont &font)
{
  recalTabsWidth();
  updateBreaks();
  QWidget::fontChange(font);
}

void TKTabBar::styleChange(QStyle &s)
{
  QWidget::styleChange(s);
  updatePixmap();
  repaint(false);
}

void TKTabBar::updatePixmap()
{
  static QCOORD break_array[] =
  {1,0, 0,1, 0,2, 1,3, 1,4, 2,5, 2,6, 1,7
  , 1,8, 0,9, 0,10, 1,11};
  d->pbreak = QPointArray(12, break_array);

  static QCOORD left_array[] =
  { 5,0, 4,1, 5,1, 3,2, 4,2, 5,2, 2,3, 3,3
  , 4,3, 5,3, 6,3, 7,3, 8,3, 1,4, 2,4, 3,4
  , 4,4, 5,4, 6,4, 7,4, 8,4, 0,5, 1,5, 2,5
  , 3,5, 4,5, 5,5, 6,5, 7,5, 8,5, 1,6, 2,6
  , 3,6, 4,6, 5,6, 6,6, 7,6, 8,6, 2,7, 3,7
  , 4,7, 5,7, 6,7, 7,7, 8,7, 3,8, 4,8, 5,8
  , 4,9, 5,9, 5,10};
  d->pleft = QPointArray(51, left_array);

  static QCOORD right_array[] =
  { 3,0, 3,1, 4,1, 3,2, 4,2, 5,2, 0,3, 1,3
  , 2,3, 3,3, 4,3, 5,3, 6,3, 0,4, 1,4, 2,4
  , 3,4, 4,4, 5,4, 6,4, 7,4, 0,5, 1,5, 2,5
  , 3,5, 4,5, 5,5, 6,5, 7,5, 8,5, 0,6, 1,6
  , 2,6, 3,6, 4,6, 5,6, 6,6, 7,6, 0,7, 1,7
  , 2,7, 3,7, 4,7, 5,7, 6,7, 3,8, 4,8, 5,8
  , 3,9, 4,9, 3,10};
  d->pright = QPointArray(51, right_array);

  static QCOORD close_array[] =
  { 0,2, 1,2, 6,2, 7,2, 1,3, 2,3, 5,3, 6,3
  , 2,4, 3,4, 4,4, 5,4, 3,5, 4,5, 2,6, 3,6
  , 4,6, 5,6, 1,7, 2,7, 5,7, 6,7, 0,8, 1,8
  , 6,8, 7,8};
  d->pclose = QPointArray(26, close_array);
}


