
#include "calendarlistview.h"
#include "calendarlistviewitem.h"
#include "calendarobject.h"

#include <vdragobject.h>
#include <vfsmimetype.h>

#include <qaccel.h>
#include <qapplication.h>
#include <qdict.h>
#include <qptrlist.h>
#include <qpopupmenu.h>

class CalendarListView::Private
{
  public:
    Private(CalendarListView*);
    virtual ~Private();
    
    VFSModule * vfs;
    int sortingColumn;
    bool ascending;
    
  private:
    CalendarListView * parent_;
};

CalendarListView::Private::Private(CalendarListView * parent)
  : vfs(0L),
    sortingColumn(0),
    ascending(true),
    parent_(parent)
{
}

CalendarListView::Private::~Private()
{
}

CalendarListView::CalendarListView(VFSModule * vfs, QWidget * parent, const char * name)
  : QListView(parent, name),
    DataHandler(vfs)
{
  d = new Private(this);
  
  d->vfs = vfs;
  
  addColumn("Name                ");
  
  setSelectionMode(Extended);
  setAllColumnsShowFocus(true);
  setShowSortIndicator(true);
  
  connect((QObject*)header(), SIGNAL(clicked(int)), this, SLOT(slotSortingColumn(int)));

/*  QAccel * accel = new QAccel(this);
  accel->connectItem
    (accel->insertItem(Key_Delete), this, SLOT(slotDelete()));*/
}

CalendarListView::~CalendarListView()
{
  if(d)
  {
    delete d;
    d = 0L;
  }
}

QDragObject * CalendarListView::dragObject()
{
  if (0 == currentItem())
    return 0;

  VDragObject * drag = new VDragObject(viewport());

//  drag->setPixmap();

  for (QListViewItemIterator it(this); it.current(); ++it)
  {
    QListViewItem * item = it.current();

    if (item->isSelected())
    {
      drag->append
        (
          static_cast<CalendarListViewItem *>(item)->path(),
          itemRect(item),
          itemRect(item)
        );
    }
  }

  return drag;
}

void CalendarListView::slotCreateVInstances(const QStringList & paths)
{
  setSorting( -1 );
        
  int i=1;
  for(QStringList::ConstIterator it=paths.begin(); it!=paths.end(); ++it)
  {
    CalendarObject *item = static_cast<CalendarObject*> (d->vfs->_fact_ref((*it), VFS_CALENDAR_MIMETYPE));
        
    if(item)
    {
      new CalendarListViewItem(d->vfs, (*it), this);
    }
          
    if( !(i%10) )
    {
      qApp->processEvents(); // Let loaded items to appear.
    }
  }
        
  setSortingColumn();
  sort();
        
  d->vfs->deleteVInstances(paths, false);      
}

void CalendarListView::slotUpdateVInstances(const QStringList & paths)
{
//  qDebug("CalendarListView::slotUpdateVInstances");
  
  QDict<char> findDict;
  
  for(QStringList::ConstIterator it=paths.begin(); it!=paths.end(); ++it)
    findDict.insert((*it), (*it).latin1());
    
  for(CalendarListViewItem *item=(CalendarListViewItem*)this->firstChild(); 
      item && findDict.size(); 
      item=(CalendarListViewItem*)item->itemBelow())
  {
    if( findDict.find( item->path() ) )
    {
      item->update();
      findDict.remove( item->path() );
    }
  }
}

void CalendarListView::slotDeleteVInstances(const QStringList & paths)
{
  qDebug("CalendarListView::slotDeleteVInstances");
  
  QDict<char> findDict;
  QPtrList<CalendarListViewItem> deletedItems;
  
  for(QStringList::ConstIterator it=paths.begin(); it!=paths.end(); ++it)
    findDict.insert((*it), (*it).latin1());
    
  for(CalendarListViewItem *item=(CalendarListViewItem*)this->firstChild(); 
      item && findDict.size(); 
      item=(CalendarListViewItem*)item->itemBelow())
  {
    if( findDict.find( item->path() ) )
    {
      findDict.remove( item->path() );
      deletedItems.append(item);
    }
  }
  
  {  
    for(QPtrListIterator<CalendarListViewItem> it(deletedItems); it.current(); ++it)
    {
      delete it.current();
    }
  }
}

void CalendarListView::slotSetFolder(const QString & folderPath)
{
  qDebug("CalendarListView::slotSetFolder %s", folderPath.latin1());
  
  // anyway, it has to be clear
  clear();
  
  VFolder * folder = vfs->folderRef(folderPath);
  if( !folder )
    return;
    
  DataHandler::setParentFolder(folder);
  
  // scan for objects
  QStringList urls = vfs->children(folder->path());
  QStringList temp;
  int i=1;
  
  if( !urls.count() )
  {
    qDebug("slotSetFolder(): No children to build.");
    return;
  }
  
  // for speeding reasons
  setSorting( -1 );
  
  for(QStringList::Iterator it=urls.begin(); it!=urls.end(); ++it)
  {
    temp.push_back(folderPath + "/" + (*it));
    if( !(i%10) )
    {
      vfs->createVInstances(temp, false);
      
      for(QStringList::Iterator it1=temp.begin(); it1!=temp.end(); ++it1)
      {
        new CalendarListViewItem(vfs, (*it1), this);
      }
      qApp->processEvents(); // Let loaded items appear.
      
      vfs->deleteVInstances(temp, false);
      
      temp.clear();
    }
    
    i++;
  }
  
  // create the last 10(or less) items
  vfs->createVInstances(temp, false);
      
  for(QStringList::Iterator it1=temp.begin(); it1!=temp.end(); ++it1)
  {
    new CalendarListViewItem(vfs, (*it1), this);
  }
      
  vfs->deleteVInstances(temp, false);
      
  temp.clear();
  
  setSortingColumn();
  sort();
}

void CalendarListView::slotCut()
{
  cutSelection();
}

void CalendarListView::slotCopy()
{
  copySelection();
}

void CalendarListView::slotPaste()
{
  paste();
}

void CalendarListView::slotDelete()
{
//  qDebug("CalendarListView::slotDelete");
  
  deleteSelection();
}

void CalendarListView::selectAll()
{
  QListView::selectAll(true);
}

void CalendarListView::unselectAll()
{
  QListView::selectAll(false);
}

bool CalendarListView::isSelectionEmpty()
{
  CalendarListViewItem * item = dynamic_cast<CalendarListViewItem*>(firstChild());
  
  while(item)
  {
    if(item->isSelected())
      return false;
      
    item = dynamic_cast<CalendarListViewItem *>(item->itemBelow());
  }
  
  return true;
}

bool CalendarListView::isSelectionSingle()
{
  CalendarListViewItem * item = dynamic_cast<CalendarListViewItem *>(firstChild());
  int iCount = 0;
    
  while(item && iCount<2)
  {
    if(item->isSelected())
      iCount++;
    
    item = dynamic_cast<CalendarListViewItem *>(item->itemBelow());
  }
  
  return iCount==1 ? true : false;
}

bool CalendarListView::isSelectionMultiple()
{
  int selCount = 0;
  
  CalendarListViewItem * item = dynamic_cast<CalendarListViewItem *>(firstChild());
  
  while(item && selCount<2)
  {
    if(item->isSelected())
      selCount++;
    item = dynamic_cast<CalendarListViewItem *>(item->itemBelow());
  }
  
  if(selCount==2)
    return true;
  
  return false;
}

QStringList CalendarListView::getSelection()
{
  QStringList selection;

  if( !isSelectionEmpty() && CALENDAR_ITEM_ID == currentItem()->rtti() )
  {
    for(CalendarListViewItem *item=(CalendarListViewItem*)this->firstChild(); item; item=(CalendarListViewItem*)item->itemBelow())
      if( item->isSelected() )
      {
        if( !item->path().isEmpty() )
          selection.append( item->path() );
      }
  }

  return selection;
}

void CalendarListView::slotSortingColumn(int column, bool ascending)
{
  d->sortingColumn = column;
  d->ascending = ascending;
}

void CalendarListView::setSortingColumn()
{
  setSorting(d->sortingColumn, d->ascending);
}

