/*
    This file is part of KOrganizer.
    Copyright (c) 2001 Cornelius Schumacher <schumacher@kde.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

//
// MailScheduler - Mail implementation of iTIP methods
//

#include <qdir.h>
#include <qfile.h>

#include <libkcal/event.h>
#include <libkcal/icalformat.h>

#include "komailclient.h"
#include "mailscheduler.h"

//QT includes
#include <qptrlist.h>
#include <qstring.h>
#include <qdir.h>
#include <qtextstream.h>
#include <qcstring.h>
#include <qfile.h>

//tkcBase includes
#include <settings.h>

// Aethera library
#include <vfsmodule.h>
#include <vfolder.h>

//KomMailPlugin includes
#include <mailagentservice.h>

#define MAILSCHEDULE_DBG

using namespace KCal;

MailScheduler::MailScheduler(Calendar *calendar)
  : IMIPScheduler(calendar),
    mailAgent( 0L ),
    vfs( 0L )
{
  qDebug( "MailScheduler::MailScheduler(Calendar *calendar)" );
  ComponentTrader* trader = dynamic_cast<ComponentTrader*>(ComponentTrader::traderRef().instance());
  if( trader )
  {
    ComponentRef ref = trader->registeredComponent( MAILAGENT_SERVICE );
    if( ref.isValid() )
       mailAgent = static_cast<MailAgentService*>( ref.instance() );
  }
  Q_CHECK_PTR( mailAgent );

  ComponentRef ref = trader->registeredComponent( AETHERA_VFS_SERVICE );
  if( ref.isValid() )
      vfs = static_cast<VFSModule*>( ref.instance() );
  Q_CHECK_PTR( vfs );
}

MailScheduler::~MailScheduler()
{
}

bool MailScheduler::publish( IncidenceBase *incidence,const QString &recipients )
{
  QString messageText = mFormat->createScheduleMessage(incidence, Scheduler::Publish);
  KOMailClient mailer;
  //  kdDebug () << "MailScheduler::publish to " << recipients << endl;
  return mailer.mailTo(incidence,recipients,messageText);
}

bool MailScheduler::performTransaction(IncidenceBase *incidence,Method method,const QString &recipients)
{
  QString messageText = mFormat->createScheduleMessage( incidence, method );
  
  KOMailClient mailer;
  if (method==Request || method==Cancel || method==Add || method==Declinecounter) {
    return mailer.mailTo(incidence,recipients,messageText);
  }
  else {
    return mailer.mailTo(incidence,recipients,messageText);
  }
}

bool MailScheduler::performTransaction(IncidenceBase *incidence, Method method)
{
  QString messageText = mFormat->createScheduleMessage(incidence,method);

  KOMailClient mailer;
  if (method==Request || method==Cancel || method==Add || method==Declinecounter || method==NoMethod)
  {
    return mailer.mailAttendees(incidence,messageText);
  }
  else
  {
    return mailer.mailOrganizer(incidence,messageText);
  }
}

QPtrList<ScheduleMessage> MailScheduler::retrieveTransactions()
{
  QPtrList<ScheduleMessage> messageList;
  MailObjectInterface * mail = 0L;
    
  QStringList urls = vfs->children( vfs->root() + "/KOPlugin-incoming" );
  for(QStringList::Iterator it=urls.begin(); it!=urls.end(); ++it)
  {
    mail = static_cast<MailObjectInterface*>
      (vfs->_fact_ref( *it, VFS_MAILS_MIMETYPE ));
    
    if( mail )
    {
      //extract the iCal part from the mail text
      // TODO: use PartObjectInterface
      QCString emailMessage( mail->message() );
      int data_begin_pos = emailMessage.find( "BEGIN:VCALENDAR" );
      int data_end_pos = emailMessage.find( "END:VCALENDAR" );
      QString messageString = emailMessage.mid( data_begin_pos, data_end_pos );
      if( messageString.isEmpty() )
        continue;
          
      ScheduleMessage *message = mFormat->parseScheduleMessage( mCalendar,messageString );
      if (message)
      {
        messageList.append(message);
        mEventMap[message->event()]= mail->path();
      }
      else
      {
        QString errorMessage;
        if (mFormat->exception())
        {
          errorMessage = mFormat->exception()->message();
          qDebug( "MailScheduler::retrieveTransactions() --> parser error: %s", errorMessage.latin1() );
        }
      }
    }
  }
  
  VFolder * folder = vfs->folderRef( vfs->root() + "/KOPlugin-incoming" );
  if( folder )
  {
    folder->freeObjects();
  }
  
  return messageList;
}

bool MailScheduler::deleteTransaction(IncidenceBase *incidence)
{
  if(  mEventMap[incidence].isEmpty()  )
    return false;
    
  mailAgent->deleteMail( mEventMap[incidence] );
  
  return true;
}
