// -*- c-basic-offset: 2 -*-
/*
   Parts of this file were part of the KDE libraries
   Copyright (c) 1999 Preston Brown <pbrown@kde.org>
   Copyright (c) 1997 Matthias Kalle Dalheimer <kalle@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
 */

#include <settings.h>

#include "kconfig.h"

QString KConfig::mAppName = "NoAppName";

class KConfig::Private
{
  public:
    QString group;
};

KConfig::KConfig()
{
  d = new Private;
  d->group = "General";
}

KConfig::~KConfig()
{
  delete d;
}

void KConfig::setAppName( const QString &appName )
{
  mAppName = appName;
}

KConfig::KConfig( const QString & filename )
{
  d = new Private;

  if (!filename.isEmpty())
  {
    qWarning("Can't do KConfig emulation with absolute path !");
    return;
  }
}

void KConfig::setGroup( const QString & group )
{
  d->group = (group.isEmpty() ? QString("General") : group);
}

QValueList<int> KConfig::readIntListEntry( const QString & pKey )
{
  QStringList strlist = readListEntry(pKey);
  QValueList<int> list;
  for (QStringList::ConstIterator it = strlist.begin(); it != strlist.end(); it++)
    // I do not check if the toInt failed because I consider the number of items
    // more important than their value
    list << (*it).toInt();

  return list;
}

int KConfig::readNumEntry( const QString & pKey, int nDefault )
{
  bool ok;
  int rc;

  QString aValue = readEntry( pKey );
  if( aValue.isNull() )
    return nDefault;
  else if( aValue == "true" )
    return 1;
  else if( aValue == "on" )
    return 1;
  else if( aValue == "yes" )
    return 1;
  else
    {
      rc = aValue.toInt( &ok );
      return( ok ? rc : nDefault );
    }
}

QString KConfig::readEntry( const QString & key, const QString &def )
{
  return Settings::instance()->readEntry("/" + mAppName + "/" + d->group + "/" + key, def);
}

QStringList KConfig::readListEntry( const QString & pKey )
{
  char sep = ',';

  QStringList list;
  QString str_list, value;
  str_list = readEntry( pKey );
  if( str_list.isEmpty() )
    return list;
  int i;
  value = "";
  int len = str_list.length();
  for( i = 0; i < len; i++ )
  {
    if( str_list[i] != sep && str_list[i] != '\\' )
    {
      value += str_list[i];
      continue;
    }
    if( str_list[i] == '\\' )
    {
      i++;
      value += str_list[i];
      continue;
    }
    list.append( value );
    value.truncate(0);
  }
  if ( str_list[len-1] != sep )
    list.append( value );
  return list;
}

bool KConfig::readBoolEntry( const QString & pKey, bool bDefault )
{
  QString aValue = readEntry( pKey );

  if( aValue.isNull() )
    return bDefault;

  else
  {
    if( aValue == "true" || aValue == "on" || aValue == "yes" || aValue == "1"
    )
      return true;
    else
    {
      bool bOK;
      int val = aValue.toInt( &bOK );
      if( bOK && val != 0 )
        return true;
      else
        return false;
    }
  }
}

QColor KConfig::readColorEntry( const QString & pKey, QColor * pDefault )
{
  QColor aRetColor;
  int nRed = 0, nGreen = 0, nBlue = 0;

  QString aValue = readEntry( pKey );
  if( !aValue.isEmpty() )
  {
    if ( aValue.at(0) == '#' )
    {
      aRetColor.setNamedColor(aValue);
    }
    else
    {

      bool bOK;

      // find first part (red)
      int nIndex = aValue.find( ',' );

      if( nIndex == -1 ){
        // return a sensible default -- Bernd
        if( pDefault )
          aRetColor = *pDefault;
        return aRetColor;
      }

      nRed = aValue.left( nIndex ).toInt( &bOK );

      // find second part (green)
      int nOldIndex = nIndex;
      nIndex = aValue.find( ',', nOldIndex+1 );

      if( nIndex == -1 ){
        // return a sensible default -- Bernd
        if( pDefault )
          aRetColor = *pDefault;
        return aRetColor;
      }
      nGreen = aValue.mid( nOldIndex+1,
        nIndex-nOldIndex-1 ).toInt( &bOK );

        // find third part (blue)
        nBlue = aValue.right( aValue.length()-nIndex-1 ).toInt( &bOK );

        aRetColor.setRgb( nRed, nGreen, nBlue );
    }
  }
  else {

    if( pDefault )
      aRetColor = *pDefault;
  }

  return aRetColor;

}

QFont KConfig::readFontEntry( const QString & pKey, QFont * pDefault )
{
  QFont aRetFont;

  QString aValue = readEntry( pKey );
  if( !aValue.isNull() ) {
    if ( aValue.contains( ',' ) > 5 ) {
      // KDE3 and upwards entry
      if ( !aRetFont.fromString( aValue ) && pDefault )
        aRetFont = *pDefault;
    }
    else {
      // backward compatibility with older font formats
      // ### remove KDE 3.1 ?
      // find first part (font family)
      int nIndex = aValue.find( ',' );
      if( nIndex == -1 ){
        if( pDefault )
          aRetFont = *pDefault;
        return aRetFont;
      }
      aRetFont.setFamily( aValue.left( nIndex ) );

      // find second part (point size)
      int nOldIndex = nIndex;
      nIndex = aValue.find( ',', nOldIndex+1 );
      if( nIndex == -1 ){
        if( pDefault )
          aRetFont = *pDefault;
        return aRetFont;
      }

      aRetFont.setPointSize( aValue.mid( nOldIndex+1,
          nIndex-nOldIndex-1 ).toInt() );

      // find third part (style hint)
      nOldIndex = nIndex;
      nIndex = aValue.find( ',', nOldIndex+1 );

      if( nIndex == -1 ){
        if( pDefault )
          aRetFont = *pDefault;
        return aRetFont;
      }

      aRetFont.setStyleHint( (QFont::StyleHint)aValue.mid( nOldIndex+1, nIndex-nOldIndex-1 ).toUInt() );

      // find fourth part (char set)
      nOldIndex = nIndex;
      nIndex = aValue.find( ',', nOldIndex+1 );

      if( nIndex == -1 ){
        if( pDefault )
          aRetFont = *pDefault;
        return aRetFont;
      }

      QString chStr=aValue.mid( nOldIndex+1,
        nIndex-nOldIndex-1 );
      // find fifth part (weight)
      nOldIndex = nIndex;
      nIndex = aValue.find( ',', nOldIndex+1 );

      if( nIndex == -1 ){
        if( pDefault )
          aRetFont = *pDefault;
        return aRetFont;
      }

      aRetFont.setWeight( aValue.mid( nOldIndex+1,
          nIndex-nOldIndex-1 ).toUInt() );

      // find sixth part (font bits)
      uint nFontBits = aValue.right( aValue.length()-nIndex-1 ).toUInt();

      aRetFont.setItalic( nFontBits & 0x01 );
      aRetFont.setUnderline( nFontBits & 0x02 );
      aRetFont.setStrikeOut( nFontBits & 0x04 );
      aRetFont.setFixedPitch( nFontBits & 0x08 );
      aRetFont.setRawMode( nFontBits & 0x20 );
    }
  }
  else
  {
    if( pDefault )
      aRetFont = *pDefault;
  }

  return aRetFont;
}

void KConfig::writeEntry( const QString & pKey, QValueList<int> list )
{
  QStringList strlist;
  QValueList<int>::ConstIterator end = list.end();
  for (QValueList<int>::ConstIterator it = list.begin(); it != end; it++)
    strlist << QString::number(*it);
  writeEntry(pKey, strlist);
}

void KConfig::writeEntry( const QString & pKey, int nValue )
{
  writeEntry( pKey, QString::number(nValue));
}

void KConfig::writeEntry( const QString & pKey, const char * s )
{
  writeEntry(pKey, QString(s));
}


void KConfig::writeEntry( const QString & pKey, const QString & s )
{
  QString k("/" + mAppName + "/" + d->group + "/" + pKey);
  Settings::instance()->writeEntry(k, s);
}

void KConfig::writeEntry( const QString & pKey, const QStringList & list )
{
  char sep = ',';

  if( list.isEmpty() )
  {
    writeEntry( pKey, QString::fromLatin1("") );
    return;
  }
  QString str_list;
  QStringList::ConstIterator it = list.begin();
  for( ; it != list.end(); ++it )
  {
    QString value = *it;
    uint i;
    for( i = 0; i < value.length(); i++ )
    {
      if( value[(int)i] == sep || value[(int)i] == '\\' )
        str_list += '\\';
      str_list += value[(int)i];
    }
    str_list += sep;
  }
  if( str_list.at(str_list.length() - 1) == sep )
    str_list.truncate( str_list.length() -1 );

  writeEntry( pKey, str_list);
}

void KConfig::writeEntry( const QString & pKey, bool bValue )
{
  QString aValue;

  if( bValue )
    aValue = "true";
  else
    aValue = "false";

  writeEntry( pKey, aValue );
}

void KConfig::writeEntry( const QString & pKey, const QColor & rColor )
{
  QString aValue;
  if (rColor.isValid())
    aValue.sprintf( "%d,%d,%d", rColor.red(), rColor.green(), rColor.blue() );
  else
    aValue = "invalid";

  writeEntry( pKey, aValue );
}

void KConfig::writeEntry( const QString & pKey, const QFont & rFont )
{
  writeEntry( pKey, rFont.toString() );
}

void KConfig::sync()
{
  // QSettings does its own syncing.
}

