#include "filedialog.h"

#include <dirs.h>

#include <qlayout.h>
#include <qstringlist.h>
#include <qlistbox.h>
#include <qdir.h>

#include <qpe/qpeapplication.h>

#include <qlineedit.h>
#include <qlabel.h>
#include <qlayout.h>

class FileDialog::Private
{
  public:
    Private();
    virtual ~Private();
  
    QVBoxLayout * mainLayout;
    QGridLayout * grid;
    
    FileSelector * fileSelector;
    
    QLabel * fileNameLabel;
    QLineEdit * fileNameEdit;
    QString defaultName;
    
    DocLnk * link;
    bool bNew;
};

FileDialog::Private::Private()
  : mainLayout(0L),
    grid(0L),
    fileSelector(0L),
    fileNameLabel(0L),
    fileNameEdit(0L),
    link(0L),
    bNew(false)
{
}

FileDialog::Private::~Private()
{
}

FileDialog::FileDialog( const QString & mimeFilter,
                        const QString & defaultName,
                        QWidget       * parent,
                        const char    * name )
  : QDialog(parent, name, true),
    d(0L)
{
  setCaption(tr("Choose document"));

  d = new Private;
  
  if( !defaultName.isEmpty() )
    d->defaultName = defaultName;
  
  d->fileSelector = new FileSelector(mimeFilter, this, "FileDialog's file selector", false, false);
  connect(d->fileSelector, SIGNAL(newSelected(const DocLnk&)), this, SLOT(slotNewSelected(const DocLnk&)));
  connect(d->fileSelector, SIGNAL(fileSelected(const DocLnk&)), this, SLOT(slotFileSelected(const DocLnk&)));
  
  d->fileNameLabel = new QLabel( tr("File Name:"), this);
  d->fileNameEdit = new QLineEdit(this);
  if( !d->defaultName.isEmpty() )
    d->fileNameEdit->setText( d->defaultName );
  d->fileNameEdit->setReadOnly(true);
  
  QVBoxLayout * layout = new QVBoxLayout(this);
  
  QGridLayout * grid = new QGridLayout(layout);
  grid->addWidget( d->fileNameLabel, 0, 0);
  grid->addWidget( d->fileNameEdit, 0, 1);

  layout->addWidget(d->fileSelector);
}

FileDialog::~FileDialog()
{
  if( d )
  {
    delete d;
    d = 0L;
  }
}

DocLnk FileDialog::getOpenFileName( const QString & mimeFilter,
                                    const QString & defaultName,
                                    QWidget       * parent,
                                    const char    * name )
{
  FileDialog docDialog(mimeFilter, defaultName, parent, name);

  docDialog.showMaximized();

  if (QDialog::Accepted != docDialog.exec())
  {
    return DocLnk();
  }

  const DocLnk * docLnk = docDialog.selected();

  if (0 == docLnk)
  {
    return DocLnk();
  }
  else
  {
    return *docLnk;
  }
}

DocLnk FileDialog::getSaveFileName( const QString & mimeFilter,
                                    const QString & defaultName,
                                    QWidget       * parent,
                                    const char    * name )
{
  FileDialog docDialog(mimeFilter, defaultName, parent, name);

  docDialog.showMaximized();
  docDialog.setNewVisible(true);

  if (QDialog::Accepted != docDialog.exec())
  {
    return DocLnk();
  }

  return docDialog.selectedSave();
}

void FileDialog::setNewVisible(bool b)
{
  if( d->fileSelector )
    d->fileSelector->setNewVisible(b);
}

void FileDialog::setCloseVisible(bool b)
{
  if( d->fileSelector )
    d->fileSelector->setCloseVisible(b);
}

const DocLnk * FileDialog::selected()
{
  return d->fileSelector->selected();
}

QString FileDialog::name(const QString & text)
{
  return QString(text);
}

QString FileDialog::desktopName(const QString & text)
{
  return QString(QPEApplication::documentDir() + text + ".desktop");
}

QString FileDialog::fileName(const QString & text)
{
  // TODO: we should autodetect the file name extension
  // using the app mime type.
  
  return QString(QPEApplication::documentDir() + text);
}

DocLnk FileDialog::selectedSave()
{
  if( d->bNew )
  {
    if( d->fileNameEdit && !d->fileNameEdit->text().isEmpty() )
    {
      QString linkName = d->fileNameEdit->text();
      
      d->link->setName( name(linkName) );
      d->link->setLinkFile( desktopName(linkName) );
      d->link->setFile( fileName(linkName) );
    }
    
    if( d->link->isValid() && d->link->writeLink() )
    {
//      qDebug("New link created:\nLinkName=%s\nFileName=%s\n", 
//        d->link->name().latin1(),
//        d->link->file().latin1() );
      
      return *d->link;
    }
    else
      return DocLnk();
  }
  else
  {
    const DocLnk * docLnk = selected();
  
    if (0 == docLnk)
    {
      return DocLnk();
    }
    else
    {
      return *docLnk;
    }
  }
}
  
void FileDialog::slotFileSelected(const DocLnk & selectedLink)
{
  if( d->fileNameEdit && selectedLink.isValid() )
  {
    d->fileNameEdit->setText( selectedLink.name() );
  }
}

void FileDialog::slotNewSelected(const DocLnk & copy)
{
  if( d->link )
  {
    delete d->link;
    d->link = 0L;
  }
  
  d->link = new DocLnk( copy );
  
  d->bNew = true;
  
  d->link->setType( AppMIME );
  
  if( d->fileNameEdit )
  {
    d->fileNameEdit->setReadOnly(false);
    d->fileNameEdit->setFocus();
  }
}

void FileDialog::slotCloseMe()
{
}

