/*
    Copyright (C) 2001-2002 by theKompany.com <www.thekompany.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    In addition, as a special exception, theKompany.com gives permission
    to link the code of this program with the tkwidgets library (or with
    modified versions of tkwidgets that use the same license as tkwidgets),
    and distribute linked combinations including the two.  You must obey
    the GNU General Public License in all respects for all of the code used
    other than tkwidgets.  If you modify this file, you may extend this
    exception to your version of the file, but you are not obligated to do so.
    If you do not wish to do so, delete this exception statement from your
    version.

    This license grants you the ability to use tkwidgets with Aethera only
    and may not be used outside of Aethera.
    See also http://www.thekompany.com/products/license.txt for details.
---------------------------------------------------------------------------------
  Date     : July 2003
  Author   : Constantinescu Eugen <eug@thekompany.com> 
  Author   : Bodi Vlad <vlad@eco-spaces.com.ro> 
  Author   : Ghiciuc Ilie <ilie@eco-spaces.com.ro>
---------------------------------------------------------------------------------
*/

//own includes
#include "filebrowser.h"
#include "filebrowserparser.h"

//QT includes
#include <qdom.h>
#include <qbuffer.h>

//standard library includes
#include <stdlib.h>

/*
----------------------------------------------------------------------------------------
FileBrowserPrivate class
----------------------------------------------------------------------------------------
*/
class FileBrowserPrivate
{
  public:
    QString generateFileName( QString& );

    QString hostAddress;
    QDomDocument doc;          
    
    WebDav* webDav;
  
    FileBrowserRequestList requestList;  
};


QString FileBrowserPrivate::generateFileName( QString& fileName)
{
  static const char letters[]="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
  
  QString prefix = "dav_";
  QString postfix = ".tmp";
  QString XXXXXX = "XXXXXX";
  
  srand( QTime::currentTime().msec() );
  int value;
  for( int count = 0; count<6; ++count )
  {
    value = rand();
    XXXXXX[count] = letters[value % 62];
  }
  
  fileName = prefix + XXXXXX + postfix;
  
  return fileName;
}


/*
----------------------------------------------------------------------------------------
FileBrowser class
----------------------------------------------------------------------------------------
*/

FileBrowser::FileBrowser( QObject* parent, const char* name )
  : QObject( parent, name),
    d( new FileBrowserPrivate )
{
  init();
  
  d->hostAddress = "unknowed";
  d->webDav = 0L;
}


FileBrowser::FileBrowser( const QString& host, QObject* parent, const char* name )
  : QObject( parent, name),
    d( new FileBrowserPrivate )
{
  init();
  
  d->hostAddress = host;
  d->webDav = 0L;
}

void FileBrowser::init()
{
  if( !d->webDav )
  {
    d->webDav = new WebDav( this, "webDavClient" );
    d->webDav->setHost( d->hostAddress );
  }
  connect( d->webDav,
           SIGNAL( requestFinished(int,bool) ),
           this,
           SLOT( slotRequestFinished(int,bool) ) 
         );
  connect( d->webDav,
           SIGNAL( readyRead(const QHttpResponseHeader&) ),
           this,
           SLOT( slotReadyRead(const QHttpResponseHeader&) )
         );
}


FileBrowser::~FileBrowser()
{
  if( !d )
  {
    delete d;
    d = 0L;
  }
}


void FileBrowser::slotReadyRead( const QHttpResponseHeader& resp )
{
  emit dataReceived( resp );
}


void FileBrowser::slotRequestFinished( int reqId, bool )
{
  FileBrowserRequestList::iterator it = d->requestList.begin();
  FileBrowserRequest request;
  for( ; it!=d->requestList.end(); ++it )
    if( (*it).getId() == reqId )
    {
      request = *it;
      break;
    }

  if( it != d->requestList.end() )
    d->requestList.removeRequest( it );
  
  if( !d->webDav->bytesAvailable() )
  {
    qDebug( "FileBrowser::slotRequestFinished() --> no bytes availabile" );
    return;
  }
    
  if( request.getOperation() == DAV_CAT )
  {
    if( !request.getIODevice()->isOpen() )
       request.getIODevice()->open( IO_ReadWrite );
    
    request.getIODevice()->writeBlock( d->webDav->readAll() );
    
    emit catFinished( reqId );
  } 
  else if( request.getOperation() == DAV_DEL ) 
  {
    emit delFinished( reqId );
  }
  else if( request.getOperation() == DAV_CREATE_FILE )
  {
    emit createFileFinished( reqId );
  }
  else if( request.getOperation() == DAV_UPLOAD )
  {
    emit uploadFinished( reqId );
  }
  else if( request.getOperation() == DAV_MKDIR )
  {
    emit mkdirFinished( reqId );
  }
  else if( request.getOperation() == DAV_LS )
  { 
    QString str( d->webDav->readAll() );

    DavFSEntityList fileList;
    request.getParser()->parseLs( str, fileList );
    
    emit lsFinished( reqId, fileList );
  }
  else if( request.getOperation() == DAV_FLAT_LS )
  { 
    QString str( d->webDav->readAll() );
    DavFSEntityList fileList;
    request.getParser()->parseFlatLs( str, fileList );

    emit flatLsFinished( reqId, fileList );
  }
  else if( request.getOperation() == DAV_FIND )
  { 
    QString str( d->webDav->readAll() );
    DavFSEntityList fileList;
    request.getParser()->parseFlatLs( str, fileList );
    
    bool result = false;
    QString seekName = request.getOptionalData( "Seek file" );
    DavFSEntityListIterator it = fileList.begin();
    for( ; it!=fileList.end(); ++it )
      if( (*it).name == seekName )
      {
        result = true;
        break;
      }
      
    emit findFinished( reqId, result );
  }
}


void FileBrowser::setHost( const QString& host )
{
  d->hostAddress = host;
}


QString FileBrowser::getHost()
{
  return d->hostAddress;  
}


void FileBrowser::setDavTransport( WebDav* transport )
{
  d->webDav = transport;
}
    
    
WebDav* FileBrowser::getDavTransport( )
{
  return d->webDav;
}


int FileBrowser::cat( const QString& path, QIODevice* data )
{
  WebDavRequestHeader request( "GET", path );
  request.setValue( "Host", d->hostAddress );
  request.setContentType( "text/xml" );
  
  int id = d->webDav->request( request );
  d->requestList.addRequest( FileBrowserRequest(id, DAV_CAT, data)  );

  return id;
}


int FileBrowser::ls( const QString& path )
{
  WebDavRequestHeader request( "PROPFIND", path );
  request.setValue( "Host", d->hostAddress + "\r\n" );
  request.setValue("DAV", "1");
  request.setValue( "Depth","1" );
  request.setContentType( "text/xml" );
  
  QDomDocument doc;

  QDomElement root = doc.createElementNS( "DAV:", "propfind" );
  root.setPrefix( "D" );  
  doc.appendChild( root );
  QDomElement prop = doc.createElement( "D:prop" );
  root.appendChild( prop );
  QDomElement firstProp = doc.createElementNS( "DAV:", "resourcetype" );
  firstProp.setPrefix( "R" );
  prop.appendChild( firstProp );
  
  QCString data = doc.toString().utf8();
  
  int id = d->webDav->request( request, data );
  
  FileBrowserRequest requestStruct(id, DAV_LS);
  FileBrowserParser* parser = new FileBrowserParser;
  requestStruct.setParser( parser );  
  d->requestList.addRequest( requestStruct );
  
  return id;
}


int FileBrowser::flat_ls( const QString& path, bool )
{
  WebDavRequestHeader request( "PROPFIND", path );
  request.setValue( "Host", d->hostAddress + "\r\n" );
  request.setValue("DAV", "1");
  request.setValue( "Depth","1" );
  request.setContentType( "text/xml" );
  
  QDomDocument doc;
  QDomElement root = doc.createElementNS( "DAV:", "propfind" );
  root.setPrefix( "D" );  
  doc.appendChild( root );
  QDomElement prop = doc.createElement( "D:prop" );
  root.appendChild( prop );
  QDomElement firstProp = doc.createElementNS( "DAV:", "resourcetype" );
  firstProp.setPrefix( "R" );
  prop.appendChild( firstProp );
  
  QCString data = doc.toString().utf8();
  
  int id = d->webDav->request( request, data );

  FileBrowserRequest requestStruct( id, DAV_FLAT_LS );
  FileBrowserParser* parser = new FileBrowserParser;
  requestStruct.setParser( parser );
  d->requestList.addRequest( requestStruct );

  return id;
}



int FileBrowser::find( const QString& path, const QString& file, bool )
{
  WebDavRequestHeader request( "PROPFIND", path );
  request.setValue( "Host", d->hostAddress + "\r\n" );
  request.setValue("DAV", "1");
  request.setValue( "Depth","1" );
  request.setContentType( "text/xml" );
  
  QDomDocument doc;
  QDomElement root = doc.createElementNS( "DAV:", "propfind" );
  root.setPrefix( "D" );  
  doc.appendChild( root );
  QDomElement prop = doc.createElement( "D:prop" );
  root.appendChild( prop );
  QDomElement firstProp = doc.createElementNS( "DAV:", "resourcetype" );
  firstProp.setPrefix( "R" );
  prop.appendChild( firstProp );
  
  QCString data = doc.toString().utf8();
  
  int id = d->webDav->request( request, data );

  FileBrowserRequest requestStruct( id, DAV_FIND );
  FileBrowserParser* parser = new FileBrowserParser;
  requestStruct.setParser( parser );
  requestStruct.setOptionalData( "Seek file", file );
  d->requestList.addRequest( requestStruct );

  return id;
}


int FileBrowser::createFile( const QString& path, QIODevice* data )
{
  WebDavRequestHeader request( "PUT", path );
  request.setValue( "Host", d->hostAddress );
  request.setContentType( "text/xml" );

  if( !data )
  {
    QCString emptyString("");
    data = new QBuffer( emptyString );
  }

  int id = d->webDav->request( request, data );
  d->requestList.addRequest( FileBrowserRequest(id, DAV_CREATE_FILE)  );

  return id;
}


int FileBrowser::upload( QString& path, QIODevice* data )
{
  if( path.isEmpty() )
    d->generateFileName( path );

  WebDavRequestHeader request( "PUT", '/' + path );
  request.setValue( "Host", d->hostAddress );
  request.setContentType( "text/xml" );

  if( !data )
  {
    QCString emptyString("");
    data = new QBuffer( emptyString );
  }

  int id = d->webDav->request( request, data );
  d->requestList.addRequest( FileBrowserRequest(id, DAV_UPLOAD)  );

  return id;
}


int FileBrowser::del( const QString& path )
{
  WebDavRequestHeader request( "DELETE", path );
  request.setValue( "Host", d->hostAddress );
  request.setContentType( "text/xml" );

  int id = d->webDav->request( request );
  d->requestList.addRequest( FileBrowserRequest(id, DAV_DEL)  );

  return id;
}


int FileBrowser::mkdir( const QString& path )
{
  WebDavRequestHeader request( "MKCOL", path );
  request.setValue( "Host", d->hostAddress );
  request.setContentType( "text/xml" );

  int id = d->webDav->request( request );
  d->requestList.addRequest( FileBrowserRequest(id, DAV_MKDIR)  );


  return id;
}

